import pytest
import ast

from vitessce import (  # noqa: F401
    VitessceConfig,
    CoordinationType as ct,
    ViewType as cm,
    DataType as dt,
    FileType as ft,
    hconcat,
    vconcat,
    AbstractWrapper,
    make_repr,
    CoordinationLevel as CL,
    AnnDataWrapper,

    # Neither of these is in the source code, but they do appear in code which is eval'd.
    VitessceChainableConfig,
    VitessceConfigDatasetFile,
)


class MockArtifactPath:
    def __init__(self, url):
        self.url = url

    def to_url(self):
        return self.url


class MockArtifact:
    def __init__(self, name, url):
        self.name = name
        self.path = MockArtifactPath(url)


def test_config_creation():
    vc = VitessceConfig(schema_version="1.0.15")
    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [],
        "coordinationSpace": {},
        "layout": [],
        "initStrategy": "auto"
    }


def test_config_add_dataset():
    vc = VitessceConfig(schema_version="1.0.15")
    vc.add_dataset(name='My Dataset')

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [],
        "initStrategy": "auto"
    }


def test_config_add_anndata_url():
    vc = VitessceConfig(schema_version="1.0.15")
    vc.add_dataset(name='My Dataset').add_object(
        AnnDataWrapper(
            adata_url="http://example.com/adata.h5ad.zarr",
            obs_set_paths=["obs/louvain"],
        )
    )

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': [
                    {
                        "fileType": "anndata.zarr",
                        "url": "http://example.com/adata.h5ad.zarr",
                        "options": {
                            "obsSets": [
                                {
                                    "name": "louvain",
                                    "path": "obs/louvain",
                                }
                            ]
                        }
                    }
                ]
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [],
        "initStrategy": "auto"
    }


def test_config_add_anndata_artifact():
    vc = VitessceConfig(schema_version="1.0.15")
    vc.add_dataset(name='My Dataset').add_object(
        AnnDataWrapper(
            adata_artifact=MockArtifact("My anndata artifact", "http://example.com/adata.h5ad.zarr"),
            obs_set_paths=["obs/louvain"],
        )
    )

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': [
                    {
                        "fileType": "anndata.zarr",
                        "url": "http://example.com/adata.h5ad.zarr",
                        "options": {
                            "obsSets": [
                                {
                                    "name": "louvain",
                                    "path": "obs/louvain",
                                }
                            ]
                        }
                    }
                ]
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [],
        "initStrategy": "auto"
    }

    vc_artifacts = vc.get_artifacts()
    assert list(vc_artifacts.keys()) == ["http://example.com/adata.h5ad.zarr"]
    assert vc_artifacts["http://example.com/adata.h5ad.zarr"].name == "My anndata artifact"


def test_config_add_dataset_add_files():
    vc = VitessceConfig(schema_version="1.0.15")
    vc.add_dataset(name='My Chained Dataset').add_file(
        url="http://example.com/cells.json",
        file_type=ft.CELLS_JSON,
        coordination_values={"obsType": "cell"},
    ).add_file(
        url="http://example.com/cell_sets.json",
        file_type=ft.CELL_SETS_JSON,
        coordination_values={"obsType": "cell"},
    )

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Chained Dataset',
                'files': [
                    {
                        'url': 'http://example.com/cells.json',
                        'fileType': 'cells.json',
                        'coordinationValues': {
                            'obsType': 'cell'
                        }
                    },
                    {
                        'url': 'http://example.com/cell_sets.json',
                        'fileType': 'cell-sets.json',
                        'coordinationValues': {
                            'obsType': 'cell'
                        }
                    }
                ]
            },
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [],
        "initStrategy": "auto"
    }


def test_config_add_spatial_view():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')

    vc.add_view(cm.SPATIAL, dataset=my_dataset)

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'h': 1,
                'w': 1,
                'x': 0,
                'y': 0
            }
        ],
        "initStrategy": "auto"
    }


def test_config_add_scatterplot_view_with_mapping():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')

    vc.add_view(
        cm.SCATTERPLOT, dataset=my_dataset, mapping="X_umap")

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
            'embeddingType': {
                'A': 'X_umap'
            }
        },
        "layout": [
            {
                'component': 'scatterplot',
                'coordinationScopes': {
                    'dataset': 'A',
                    'embeddingType': 'A'
                },
                'h': 1,
                'w': 1,
                'x': 0,
                'y': 0
            }
        ],
        "initStrategy": "auto"
    }


def test_config_add_scatterplot_view_with_embedding_coordinations():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')

    my_view = vc.add_view(cm.SCATTERPLOT, dataset=my_dataset)

    et_scope, ez_scope, ex_scope, ey_scope = vc.add_coordination(
        ct.EMBEDDING_TYPE, ct.EMBEDDING_ZOOM, ct.EMBEDDING_TARGET_X, ct.EMBEDDING_TARGET_Y)
    my_view.use_coordination(et_scope, ez_scope, ex_scope, ey_scope)

    et_scope.set_value("X_pca")
    ez_scope.set_value(2)
    ex_scope.set_value(10)
    ey_scope.set_value(11)

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
            'embeddingType': {
                'A': 'X_pca'
            },
            'embeddingZoom': {
                'A': 2
            },
            'embeddingTargetX': {
                'A': 10
            },
            'embeddingTargetY': {
                'A': 11
            },
        },
        "layout": [
            {
                'component': 'scatterplot',
                'coordinationScopes': {
                    'dataset': 'A',
                    'embeddingType': 'A',
                    'embeddingZoom': 'A',
                    'embeddingTargetX': 'A',
                    'embeddingTargetY': 'A',
                },
                'h': 1,
                'w': 1,
                'x': 0,
                'y': 0
            }
        ],
        "initStrategy": "auto"
    }


def test_config_add_dataset_add_objects():
    vc = VitessceConfig(schema_version="1.0.15")

    class MockWrapperA(AbstractWrapper):
        def __init__(self, name, **kwargs):
            super().__init__(**kwargs)
            self.name = name

        def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
            def get_molecules(base_url):
                return {
                    "url": f"{base_url}/molecules",
                    "fileType": "molecules.json",
                    "coordinationValues": {
                        "obsType": "molecule"
                    }
                }

            def get_cells(base_url):
                return {
                    "url": f"{base_url}/cells",
                    "fileType": "cells.json"
                }
            self.file_def_creators += [get_molecules, get_cells]

    class MockWrapperB(AbstractWrapper):
        def __init__(self, name, **kwargs):
            super().__init__(**kwargs)
            self.name = name

        def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
            def get_cell_sets(base_url):
                return {
                    "url": f"{base_url}/cell-sets",
                    "fileType": "cell-sets.json"
                }
            self.file_def_creators += [get_cell_sets]

    vc.add_dataset(name='My Object Dataset').add_object(
        obj=MockWrapperA("Experiment A")
    ).add_object(
        obj=MockWrapperB("Experiment B")
    )

    vc_dict = vc.to_dict(base_url="http://localhost:8000")

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Object Dataset',
                'files': [
                    {
                        "url": "http://localhost:8000/molecules",
                        "fileType": "molecules.json",
                        "coordinationValues": {
                            "obsType": "molecule"
                        }
                    },
                    {
                        "url": "http://localhost:8000/cells",
                        "fileType": "cells.json"
                    },
                    {
                        "url": "http://localhost:8000/cell-sets",
                        "fileType": "cell-sets.json"
                    }
                ]
            },
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [],
        "initStrategy": "auto"
    }


def test_config_set_layout_single_view():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')
    my_view = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    vc.layout(my_view)

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 0,
                'y': 0,
                'h': 12,
                'w': 12,
            }
        ],
        "initStrategy": "auto"
    }


def test_config_set_layout_multi_view():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')
    v1 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v2 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v3 = vc.add_view(cm.SPATIAL, dataset=my_dataset)

    vc.layout(hconcat(v1, vconcat(v2, v3)))

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 0,
                'y': 0,
                'h': 12,
                'w': 6,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 6,
                'y': 0,
                'h': 6,
                'w': 6,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 6,
                'y': 6,
                'h': 6,
                'w': 6,
            }
        ],
        "initStrategy": "auto"
    }


def test_config_set_layout_multi_view_custom_split():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')
    v1 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v2 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v3 = vc.add_view(cm.SPATIAL, dataset=my_dataset)

    vc.layout(hconcat(v1, vconcat(v2, v3, split=[1, 2]), split=[3, 1]))

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 0,
                'y': 0,
                'h': 12,
                'w': 9,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 9,
                'y': 0,
                'h': 4,
                'w': 3,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 9,
                'y': 4,
                'h': 8,
                'w': 3,
            }
        ],
        "initStrategy": "auto"
    }


def test_config_set_layout_multi_view_magic():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')
    v1 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v2 = vc.add_view(cm.SPATIAL, dataset=my_dataset)
    v3 = vc.add_view(cm.SPATIAL, dataset=my_dataset)

    vc.layout(v1 | (v2 / v3))

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "",
        "description": "",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
        },
        "layout": [
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 0,
                'y': 0,
                'h': 12,
                'w': 6,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 6,
                'y': 0,
                'h': 6,
                'w': 6,
            },
            {
                'component': 'spatial',
                'coordinationScopes': {
                    'dataset': 'A',
                },
                'x': 6,
                'y': 6,
                'h': 6,
                'w': 6,
            }
        ],
        "initStrategy": "auto"
    }


def test_config_from_dict():
    vc = VitessceConfig.from_dict({
        "version": "1.0.15",
        "name": "Test name",
        "description": "Test description",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My First Dataset',
                'files': [
                    {
                        'url': 'http://cells.json',
                        'fileType': 'cells.json',
                        'requestInit': {
                            'headers': {
                                'Authorization': 'Bearer token'
                            }
                        }
                    }
                ]
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A'
            },
            'spatialZoom': {
                'ABC': 11
            },
        },
        "layout": [
            {
                "component": "spatial",
                "props": {
                    "cellRadius": 50
                },
                "coordinationScopes": {
                    "spatialZoom": 'ABC'
                },
                "x": 5,
                "y": 0,
                "w": 4,
                "h": 4
            },
        ],
        "initStrategy": "auto"
    })

    vc.add_dataset(name='My Second Dataset')

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.15",
        "name": "Test name",
        "description": "Test description",
        "datasets": [
            {
                'uid': 'A',
                'name': 'My First Dataset',
                'files': [
                    {
                        'url': 'http://cells.json',
                        'fileType': 'cells.json',
                        'requestInit': {
                            'headers': {
                                'Authorization': 'Bearer token'
                            }
                        }
                    }
                ]
            },
            {
                'uid': 'B',
                'name': 'My Second Dataset',
                'files': []
            }
        ],
        'coordinationSpace': {
            'dataset': {
                'A': 'A',
                'B': 'B'
            },
            'spatialZoom': {
                'ABC': 11,
            },
        },
        "layout": [
            {
                "component": "spatial",
                "props": {
                    "cellRadius": 50
                },
                "coordinationScopes": {
                    "spatialZoom": 'ABC'
                },
                "x": 5,
                "y": 0,
                "w": 4,
                "h": 4
            },
        ],
        "initStrategy": "auto"
    }


def test_config_from_dict_raises_error_if_dataset_ambiguous():
    with pytest.raises(ValueError):
        VitessceConfig.from_dict({
            "version": "1.0.15",
            "name": "Test name",
            "description": "Test description",
            "datasets": [
                {
                    'uid': 'A',
                    'name': 'My First Dataset',
                    'files': [
                        {
                            'url': 'http://cells-1.json',
                            'fileType': 'cells.json'
                        }
                    ]
                },
                {
                    'uid': 'B',
                    'name': 'My Second Dataset',
                    'files': [
                        {
                            'url': 'http://cells-2.json',
                            'fileType': 'cells.json'
                        }
                    ]
                }
            ],
            'coordinationSpace': {
                'dataset': {
                    'A': 'A'
                },
                'spatialZoom': {
                    'ABC': 11
                },
            },
            "layout": [
                {
                    "component": "spatial",
                    "props": {
                        "cellRadius": 50
                    },
                    "coordinationScopes": {
                        "spatialZoom": 'ABC'
                    },
                    "x": 5,
                    "y": 0,
                    "w": 4,
                    "h": 4
                },
            ],
            "initStrategy": "auto"
        })


def test_config_to_python_with_data_objects():
    vc = VitessceConfig(schema_version="1.0.15")

    class MockWrapperA(AbstractWrapper):
        def __init__(self, name, **kwargs):
            super().__init__(**kwargs)
            self._repr = make_repr(locals())
            self.name = name

        def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
            def get_molecules(base_url):
                return {
                    "url": f"{base_url}/molecules",
                    "fileType": "molecules.json",
                    "coordinationValues": {
                        "obsType": "molecule"
                    }
                }

            def get_cells(base_url):
                return {
                    "url": f"{base_url}/cells",
                    "fileType": "cells.json"
                }
            self.file_def_creators += [get_molecules, get_cells]

    class MockWrapperB(AbstractWrapper):
        def __init__(self, name, **kwargs):
            super().__init__(**kwargs)
            self._repr = make_repr(locals())
            self.name = name

        def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
            def get_cell_sets(base_url):
                return {
                    "url": f"{base_url}/cell-sets",
                    "fileType": "cell-sets.json"
                }
            self.file_def_creators += [get_cell_sets]

    dataset_a = vc.add_dataset(name='My First Dataset').add_object(
        obj=MockWrapperA("Experiment A")
    ).add_file(
        url="http://example.com/my_cells.json",
        file_type=ft.CELLS_JSON
    )
    dataset_b = vc.add_dataset(name='My Second Dataset').add_object(
        obj=MockWrapperB("Experiment B")
    )
    vc.add_view(cm.SPATIAL, dataset=dataset_a, x=0, y=0,
                w=3, h=3).set_props(title="My spatial plot")
    vc.add_view(cm.SCATTERPLOT, dataset=dataset_b, x=3, y=0, w=3,
                h=3, mapping="PCA").set_props(title="My scatterplot")
    base_url = "http://localhost:8000"

    classes_to_import, code_block = vc.to_python()
    assert classes_to_import == ['VitessceChainableConfig', 'VitessceConfigDatasetFile']

    # Evaluate the code string directly
    reconstructed_vc = eval(code_block)
    assert vc.to_dict(base_url=base_url) == reconstructed_vc.to_dict(base_url=base_url)

    # Convert code string to an AST and back before evaluation
    if hasattr(ast, 'unparse'):  # pragma: no cover
        # Unparse added in Python 3.9
        ast_reconstructed_vc = eval(ast.unparse(ast.parse(code_block)))
        assert vc.to_dict(base_url=base_url) == ast_reconstructed_vc.to_dict(base_url=base_url)
    else:  # pragma: no cover
        ast.parse(code_block)


def test_config_add_coordination_by_dict():
    vc = VitessceConfig(schema_version="1.0.16")

    vc.add_coordination_by_dict({
        'spatialImageLayer': CL([
            {
                'image': 'S-1905-017737_bf',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialImageChannel': CL([
                    {
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'spatialTargetC': 1,
                        'spatialChannelColor': [0, 255, 0],
                    },
                ]),
            },
        ]),
        'spatialSegmentationLayer': CL([
            {
                'image': 'S-1905-017737',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialSegmentationChannel': CL([
                    {
                        'obsType': 'Cortical Interstitia',
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Non-Globally Sclerotic Glomeruli',
                        'spatialTargetC': 1,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Globally Sclerotic Glomeruli',
                        'spatialTargetC': 2,
                        'spatialChannelColor': [255, 0, 0],
                    },
                ]),
            },
        ]),
    })

    vc_dict = vc.to_dict()

    assert vc_dict["coordinationSpace"] == {
        "spatialImageLayer": {"A": '__dummy__'},
        "image": {"A": 'S-1905-017737_bf', "B": 'S-1905-017737'},
        "spatialLayerVisible": {"A": True, "B": True},
        "spatialLayerOpacity": {"A": 1, "B": 1},
        "spatialImageChannel": {"A": '__dummy__', "B": '__dummy__'},
        "spatialTargetC": {
            "A": 0, "B": 1, "C": 0, "D": 1, "E": 2,
        },
        "spatialChannelColor": {
            "A": [255, 0, 0], "B": [0, 255, 0], "C": [255, 0, 0], "D": [255, 0, 0], "E": [255, 0, 0],
        },
        "spatialSegmentationLayer": {"A": '__dummy__'},
        "spatialSegmentationChannel": {"A": '__dummy__', "B": '__dummy__', "C": '__dummy__'},
        "obsType": {
            "A": 'Cortical Interstitia',
            "B": 'Non-Globally Sclerotic Glomeruli',
            "C": 'Globally Sclerotic Glomeruli',
        },
    }


def test_config_add_and_use_coordination_by_dict():
    vc = VitessceConfig(schema_version="1.0.16", name="My config")
    dataset = vc.add_dataset(name="My dataset")

    (color_scope, ) = vc.add_coordination('spatialChannelColor')
    color_scope.set_value([255, 0, 0])

    scopes = vc.add_coordination_by_dict({
        "spatialImageLayer": CL([
            {
                "image": 'S-1905-017737_bf',
                "spatialLayerVisible": True,
                "spatialLayerOpacity": 1,
                "spatialImageChannel": CL([
                    {
                        "spatialTargetC": 0,
                        "spatialChannelColor": [0, 255, 0],
                    },
                    {
                        "spatialTargetC": 1,
                        "spatialChannelColor": [0, 0, 255],
                    },
                ]),
            },
        ]),
        "spatialSegmentationLayer": CL([
            {
                "image": 'S-1905-017737',
                "spatialLayerVisible": True,
                "spatialLayerOpacity": 1,
                "spatialSegmentationChannel": CL([
                    {
                        "obsType": 'Cortical Interstitia',
                        "spatialTargetC": 0,
                        "spatialChannelColor": color_scope,
                    },
                    {
                        "obsType": 'Non-Globally Sclerotic Glomeruli',
                        "spatialTargetC": 1,
                        "spatialChannelColor": color_scope,
                    },
                    {
                        "obsType": 'Globally Sclerotic Glomeruli',
                        "spatialTargetC": 2,
                        "spatialChannelColor": color_scope,
                    },
                ]),
            },
        ]),
    })

    spatial_view = vc.add_view('spatial', dataset=dataset)
    spatial_view.use_coordination_by_dict(scopes)

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.16",
        "name": "My config",
        "description": "",
        "datasets": [
            {
                "uid": "A",
                "name": "My dataset",
                "files": []
            }
        ],
        "coordinationSpace": {
            "dataset": {
                "A": "A"
            },
            "spatialImageLayer": {
                "A": "__dummy__"
            },
            "image": {
                "A": "S-1905-017737_bf",
                "B": "S-1905-017737"
            },
            "spatialLayerVisible": {
                "A": True,
                "B": True
            },
            "spatialLayerOpacity": {
                "A": 1,
                "B": 1
            },
            "spatialImageChannel": {
                "A": "__dummy__",
                "B": "__dummy__"
            },
            "spatialTargetC": {
                "A": 0,
                "B": 1,
                "C": 0,
                "D": 1,
                "E": 2
            },
            "spatialChannelColor": {
                "A": [255, 0, 0],
                "B": [0, 255, 0],
                "C": [0, 0, 255]
            },
            "spatialSegmentationLayer": {
                "A": "__dummy__"
            },
            "spatialSegmentationChannel": {
                "A": "__dummy__",
                "B": "__dummy__",
                "C": "__dummy__"
            },
            "obsType": {
                "A": "Cortical Interstitia",
                "B": "Non-Globally Sclerotic Glomeruli",
                "C": "Globally Sclerotic Glomeruli"
            }
        },
        "layout": [
            {
                "component": "spatial",
                "coordinationScopes": {
                    "spatialImageLayer": ["A"],
                    "spatialSegmentationLayer": ["A"]
                },
                "coordinationScopesBy": {
                    "spatialImageLayer": {
                        "image": {
                            "A": "A"
                        },
                        "spatialLayerVisible": {
                            "A": "A"
                        },
                        "spatialLayerOpacity": {
                            "A": "A"
                        },
                        "spatialImageChannel": {
                            "A": ["A", "B"]
                        }
                    },
                    "spatialImageChannel": {
                        "spatialTargetC": {
                            "A": "A",
                            "B": "B"
                        },
                        "spatialChannelColor": {
                            "A": "B",
                            "B": "C"
                        }
                    },
                    "spatialSegmentationLayer": {
                        "image": {
                            "A": "B"
                        },
                        "spatialLayerVisible": {
                            "A": "B"
                        },
                        "spatialLayerOpacity": {
                            "A": "B"
                        },
                        "spatialSegmentationChannel": {
                            "A": ["A", "B", "C"]
                        }
                    },
                    "spatialSegmentationChannel": {
                        "obsType": {
                            "A": "A",
                            "B": "B",
                            "C": "C"
                        },
                        "spatialTargetC": {
                            "A": "C",
                            "B": "D",
                            "C": "E"
                        },
                        "spatialChannelColor": {
                            "A": "A",
                            "B": "A",
                            "C": "A"
                        }
                    }
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            }
        ],
        "initStrategy": "auto"
    }


def test_config_use_meta_complex_coordination():
    vc = VitessceConfig(schema_version="1.0.16", name="My config")
    dataset = vc.add_dataset(name="My dataset")

    scopes = vc.add_coordination_by_dict({
        'spatialImageLayer': CL([
            {
                'image': 'S-1905-017737_bf',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialImageChannel': CL([
                    {
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'spatialTargetC': 1,
                        'spatialChannelColor': [0, 255, 0],
                    },
                ]),
            },
        ]),
        'spatialSegmentationLayer': CL([
            {
                'image': 'S-1905-017737',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialSegmentationChannel': CL([
                    {
                        'obsType': 'Cortical Interstitia',
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Non-Globally Sclerotic Glomeruli',
                        'spatialTargetC': 1,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Globally Sclerotic Glomeruli',
                        'spatialTargetC': 2,
                        'spatialChannelColor': [255, 0, 0],
                    },
                ]),
            },
        ]),
    })

    meta_coordination_scope = vc.add_meta_coordination()
    meta_coordination_scope.use_coordination_by_dict(scopes)

    spatial_view = vc.add_view('spatial', dataset=dataset)
    lc_view = vc.add_view('layerController', dataset=dataset)

    spatial_view.use_meta_coordination(meta_coordination_scope)
    lc_view.use_meta_coordination(meta_coordination_scope)

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.16",
        "name": "My config",
        "description": "",
        "datasets": [
            {
                "uid": "A",
                "name": "My dataset",
                "files": []
            }
        ],
        "coordinationSpace": {
            "dataset": {
                "A": "A"
            },
            "spatialImageLayer": {
                "A": "__dummy__"
            },
            "image": {
                "A": "S-1905-017737_bf",
                "B": "S-1905-017737"
            },
            "spatialLayerVisible": {
                "A": True,
                "B": True
            },
            "spatialLayerOpacity": {
                "A": 1,
                "B": 1
            },
            "spatialImageChannel": {
                "A": "__dummy__",
                "B": "__dummy__"
            },
            "spatialTargetC": {
                "A": 0,
                "B": 1,
                "C": 0,
                "D": 1,
                "E": 2
            },
            "spatialChannelColor": {
                "A": [255, 0, 0],
                "B": [0, 255, 0],
                "C": [255, 0, 0],
                "D": [255, 0, 0],
                "E": [255, 0, 0]
            },
            "spatialSegmentationLayer": {
                "A": "__dummy__"
            },
            "spatialSegmentationChannel": {
                "A": "__dummy__",
                "B": "__dummy__",
                "C": "__dummy__"
            },
            "obsType": {
                "A": "Cortical Interstitia",
                "B": "Non-Globally Sclerotic Glomeruli",
                "C": "Globally Sclerotic Glomeruli"
            },
            "metaCoordinationScopes": {
                "A": {
                    "spatialImageLayer": ["A"],
                    "spatialSegmentationLayer": ["A"]
                }
            },
            "metaCoordinationScopesBy": {
                "A": {
                    "spatialImageLayer": {
                        "image": {
                            "A": "A"
                        },
                        "spatialLayerVisible": {
                            "A": "A"
                        },
                        "spatialLayerOpacity": {
                            "A": "A"
                        },
                        "spatialImageChannel": {
                            "A": ["A", "B"]
                        }
                    },
                    "spatialImageChannel": {
                        "spatialTargetC": {
                            "A": "A",
                            "B": "B"
                        },
                        "spatialChannelColor": {
                            "A": "A",
                            "B": "B"
                        }
                    },
                    "spatialSegmentationLayer": {
                        "image": {
                            "A": "B"
                        },
                        "spatialLayerVisible": {
                            "A": "B"
                        },
                        "spatialLayerOpacity": {
                            "A": "B"
                        },
                        "spatialSegmentationChannel": {
                            "A": ["A", "B", "C"]
                        }
                    },
                    "spatialSegmentationChannel": {
                        "obsType": {
                            "A": "A",
                            "B": "B",
                            "C": "C"
                        },
                        "spatialTargetC": {
                            "A": "C",
                            "B": "D",
                            "C": "E"
                        },
                        "spatialChannelColor": {
                            "A": "C",
                            "B": "D",
                            "C": "E"
                        }
                    }
                }
            }
        },
        "layout": [
            {
                "component": "spatial",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["A"],
                    "metaCoordinationScopesBy": ["A"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            },
            {
                "component": "layerController",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["A"],
                    "metaCoordinationScopesBy": ["A"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            }
        ],
        "initStrategy": "auto"
    }


def test_config_link_views_by_dict():
    vc = VitessceConfig(schema_version="1.0.16", name="My config")
    dataset = vc.add_dataset(name="My dataset")

    spatial_view = vc.add_view('spatial', dataset=dataset)
    lc_view = vc.add_view('layerController', dataset=dataset)

    vc.link_views_by_dict([spatial_view, lc_view], {
        'spatialImageLayer': CL([
            {
                'image': 'S-1905-017737_bf',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialImageChannel': CL([
                    {
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'spatialTargetC': 1,
                        'spatialChannelColor': [0, 255, 0],
                    },
                ]),
            },
        ]),
        'spatialSegmentationLayer': CL([
            {
                'image': 'S-1905-017737',
                'spatialLayerVisible': True,
                'spatialLayerOpacity': 1,
                'spatialSegmentationChannel': CL([
                    {
                        'obsType': 'Cortical Interstitia',
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Non-Globally Sclerotic Glomeruli',
                        'spatialTargetC': 1,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'obsType': 'Globally Sclerotic Glomeruli',
                        'spatialTargetC': 2,
                        'spatialChannelColor': [255, 0, 0],
                    },
                ]),
            },
        ]),
    })

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.16",
        "name": "My config",
        "description": "",
        "datasets": [
            {
                "uid": "A",
                "name": "My dataset",
                "files": []
            }
        ],
        "coordinationSpace": {
            "dataset": {
                "A": "A"
            },
            "spatialImageLayer": {
                "A": "__dummy__"
            },
            "image": {
                "A": "S-1905-017737_bf",
                "B": "S-1905-017737"
            },
            "spatialLayerVisible": {
                "A": True,
                "B": True
            },
            "spatialLayerOpacity": {
                "A": 1,
                "B": 1
            },
            "spatialImageChannel": {
                "A": "__dummy__",
                "B": "__dummy__"
            },
            "spatialTargetC": {
                "A": 0,
                "B": 1,
                "C": 0,
                "D": 1,
                "E": 2
            },
            "spatialChannelColor": {
                "A": [255, 0, 0],
                "B": [0, 255, 0],
                "C": [255, 0, 0],
                "D": [255, 0, 0],
                "E": [255, 0, 0]
            },
            "spatialSegmentationLayer": {
                "A": "__dummy__"
            },
            "spatialSegmentationChannel": {
                "A": "__dummy__",
                "B": "__dummy__",
                "C": "__dummy__"
            },
            "obsType": {
                "A": "Cortical Interstitia",
                "B": "Non-Globally Sclerotic Glomeruli",
                "C": "Globally Sclerotic Glomeruli"
            },
            "metaCoordinationScopes": {
                "A": {
                    "spatialImageLayer": ["A"],
                    "spatialSegmentationLayer": ["A"]
                }
            },
            "metaCoordinationScopesBy": {
                "A": {
                    "spatialImageLayer": {
                        "image": {
                            "A": "A"
                        },
                        "spatialLayerVisible": {
                            "A": "A"
                        },
                        "spatialLayerOpacity": {
                            "A": "A"
                        },
                        "spatialImageChannel": {
                            "A": ["A", "B"]
                        }
                    },
                    "spatialImageChannel": {
                        "spatialTargetC": {
                            "A": "A",
                            "B": "B"
                        },
                        "spatialChannelColor": {
                            "A": "A",
                            "B": "B"
                        }
                    },
                    "spatialSegmentationLayer": {
                        "image": {
                            "A": "B"
                        },
                        "spatialLayerVisible": {
                            "A": "B"
                        },
                        "spatialLayerOpacity": {
                            "A": "B"
                        },
                        "spatialSegmentationChannel": {
                            "A": ["A", "B", "C"]
                        }
                    },
                    "spatialSegmentationChannel": {
                        "obsType": {
                            "A": "A",
                            "B": "B",
                            "C": "C"
                        },
                        "spatialTargetC": {
                            "A": "C",
                            "B": "D",
                            "C": "E"
                        },
                        "spatialChannelColor": {
                            "A": "C",
                            "B": "D",
                            "C": "E"
                        }
                    }
                }
            }
        },
        "layout": [
            {
                "component": "spatial",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["A"],
                    "metaCoordinationScopesBy": ["A"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            },
            {
                "component": "layerController",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["A"],
                    "metaCoordinationScopesBy": ["A"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            }
        ],
        "initStrategy": "auto"
    }


def test_config_link_views_by_dict_with_scope_prefix():
    vc = VitessceConfig(schema_version="1.0.16", name="My config")
    dataset = vc.add_dataset(name="My dataset")

    spatial_view = vc.add_view('spatial', dataset=dataset)
    lc_view = vc.add_view('layerController', dataset=dataset)

    vc.link_views_by_dict([spatial_view, lc_view], {
        'spatialImageLayer': CL([
            {
                'spatialLayerOpacity': 1,
                'spatialImageChannel': CL([
                    {
                        'spatialTargetC': 0,
                        'spatialChannelColor': [255, 0, 0],
                    },
                    {
                        'spatialTargetC': 1,
                        'spatialChannelColor': [0, 255, 0],
                    },
                ]),
            },
        ])
    }, scope_prefix="SOME_PREFIX_")

    vc_dict = vc.to_dict()

    assert vc_dict == {
        "version": "1.0.16",
        "name": "My config",
        "description": "",
        "datasets": [
            {
                "uid": "A",
                "name": "My dataset",
                "files": []
            }
        ],
        "coordinationSpace": {
            "dataset": {
                "A": "A"
            },
            "spatialImageLayer": {
                "SOME_PREFIX_0": "__dummy__"
            },
            "spatialLayerOpacity": {
                "SOME_PREFIX_0": 1,
            },
            "spatialImageChannel": {
                "SOME_PREFIX_0": "__dummy__",
                "SOME_PREFIX_1": "__dummy__"
            },
            "spatialTargetC": {
                "SOME_PREFIX_0": 0,
                "SOME_PREFIX_1": 1
            },
            "spatialChannelColor": {
                "SOME_PREFIX_0": [255, 0, 0],
                "SOME_PREFIX_1": [0, 255, 0],
            },
            "metaCoordinationScopes": {
                "SOME_PREFIX_0": {
                    "spatialImageLayer": ["SOME_PREFIX_0"]
                }
            },
            "metaCoordinationScopesBy": {
                "SOME_PREFIX_0": {
                    "spatialImageLayer": {
                        "spatialLayerOpacity": {
                            "SOME_PREFIX_0": "SOME_PREFIX_0"
                        },
                        "spatialImageChannel": {
                            "SOME_PREFIX_0": ["SOME_PREFIX_0", "SOME_PREFIX_1"]
                        }
                    },
                    "spatialImageChannel": {
                        "spatialTargetC": {
                            "SOME_PREFIX_0": "SOME_PREFIX_0",
                            "SOME_PREFIX_1": "SOME_PREFIX_1"
                        },
                        "spatialChannelColor": {
                            "SOME_PREFIX_0": "SOME_PREFIX_0",
                            "SOME_PREFIX_1": "SOME_PREFIX_1"
                        }
                    }
                }
            }
        },
        "layout": [
            {
                "component": "spatial",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["SOME_PREFIX_0"],
                    "metaCoordinationScopesBy": ["SOME_PREFIX_0"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            },
            {
                "component": "layerController",
                "coordinationScopes": {
                    "dataset": "A",
                    "metaCoordinationScopes": ["SOME_PREFIX_0"],
                    "metaCoordinationScopesBy": ["SOME_PREFIX_0"]
                },
                "x": 0, "y": 0, "w": 1, "h": 1
            }
        ],
        "initStrategy": "auto"
    }



def get_next_scope_numeric(prev_scopes):
    next_scope_int = 0
    next_scope_str = None

    while True:
        next_scope_str = str(next_scope_int)
        if next_scope_str not in prev_scopes:
            break
        next_scope_int += 1
    return next_scope_str


def create_prefixed_get_next_scope_numeric(prefix):

    def inner_get_next_scope(prev_scopes):
        next_scope_int = 0
        next_scope_str = None

        while True:
            next_scope_str = f"{prefix}{next_scope_int}"
            if next_scope_str not in prev_scopes:
                break
            next_scope_int += 1
        return next_scope_str

    return inner_get_next_scope


def get_initial_coordination_scope_prefix(dataset_uid, data_type):
    return f"init_{dataset_uid}_{data_type}_"


def get_initial_coordination_scope_name(dataset_uid, data_type, i=None):
    prefix = get_initial_coordination_scope_prefix(dataset_uid, data_type)
    return f"{prefix}{0 if i is None else i}"



import pytest

from vitessce.data_utils.entities import (
    CellSets,
    Cells,
)


def test_cells():

    cell_ids = ['cell_1', 'cell_2', 'cell_3']

    cells = Cells(cell_ids=cell_ids)
    assert list(cells.json.keys()) == cell_ids

    cells.add_mapping('umap', [[1, 1], [2, 2], [3, 3]])

    cells.add_mapping('pca', [[1, 1], [2, 2], [3, 3]])

    cells.add_centroids([[1, 1], [2, 2], [3, 3]])

    cells.add_polygon_outline(
        [[[1, 1], [1, 1], [1, 1]], [[2, 2], [2, 2], [2, 2]], [[3, 3], [3, 3], [3, 3]]])

    assert cells.json == {
        'cell_1': {
            'mappings': {'umap': [1, 1], 'pca': [1, 1]},
            'xy': [1, 1],
            'poly': [[1, 1], [1, 1], [1, 1]]
        },
        'cell_2': {
            'mappings': {'umap': [2, 2], 'pca': [2, 2]},
            'xy': [2, 2],
            'poly': [[2, 2], [2, 2], [2, 2]]
        },
        'cell_3': {
            'mappings': {'umap': [3, 3], 'pca': [3, 3]},
            'xy': [3, 3],
            'poly': [[3, 3], [3, 3], [3, 3]]
        }
    }


def test_cells_bad_polygon_outline_type():

    cell_ids = ['cell_1', 'cell_2', 'cell_3']
    cells = Cells(cell_ids=cell_ids)
    with pytest.raises(Exception) as context:
        # The extra 3 should be problematic since polygons are two dimensional.
        cells.add_polygon_outline([
            [[1, 1, 3], [1, 1], [1, 1]],
            [[2, 2], [2, 2], [2, 2]],
            [[3, 3], [3, 3], [3, 3]]
        ])
    assert 'Polygon outline for cell_1 should be a list of two element lists i.e xy coordinates' in str(context)


def test_cells_bad_mappings_length():

    cell_ids = ['cell_1', 'cell_2', 'cell_3']
    cells = Cells(cell_ids=cell_ids)
    with pytest.raises(Exception) as context:
        # There are 3 cells in this object so only passing in two scatterplot cooridnates is problematic.
        cells.add_mapping('umap', [[1, 1], [2, 2]])
    assert 'Coordinates length does not match Cell IDs Length' in str(context)


def test_cell_sets():

    cell_sets = CellSets()
    cell_sets.add_level_zero_node('Clusters')

    cell_sets.add_node('Cluster 1', ['Clusters'])
    cell_sets.add_node('Cluster 2', ['Clusters'])
    cell_sets.add_node('Subcluster 1', ['Clusters', 'Cluster 1'], ['cell_1', 'cell_2'])
    cell_sets.add_node('Subcluster 2', ['Clusters', 'Cluster 1'], ['cell_3', 'cell_4'])

    assert cell_sets.json == {
        "datatype": "cell",
        "version": "0.1.2",
        "tree": [{
            "name": 'Clusters',
            "children": [
                {
                    "name": 'Cluster 1',
                    "children": [
                        {
                            "name": 'Subcluster 1',
                            "set": ['cell_1', 'cell_2']
                        },
                        {
                            "name": 'Subcluster 2',
                            "set": ['cell_3', 'cell_4']
                        }
                    ]
                },
                {
                    "name": 'Cluster 2',
                }
            ]
        }]
    }


def test_cell_sets_node_not_found():

    cell_sets = CellSets()
    with pytest.raises(Exception) as context:
        cell_sets.add_node('Cluster 1', ['Clusters Not Found'])

    assert "No node with path ['Clusters Not Found'] found to add Cluster 1 to" in str(context)



import pytest

from vitessce import (
    VitessceConfig,
    ViewType as cm,
)


@pytest.fixture
def vitessce_config():
    vc = VitessceConfig(schema_version="1.0.15")
    my_dataset = vc.add_dataset(name='My Dataset')
    vc.add_view(cm.SPATIAL, dataset=my_dataset)
    vc.add_view(cm.SCATTERPLOT, dataset=my_dataset)
    vc.add_view(cm.SCATTERPLOT, dataset=my_dataset)
    return vc


def test_get_views(vitessce_config):
    views = vitessce_config.get_views()
    assert len(views) == 3
    assert views[0].view["component"].lower() == "spatial"
    assert views[1].view["component"].lower() == "scatterplot"


def test_get_view_by_index(vitessce_config):
    view = vitessce_config.get_view_by_index(0)
    vc_dict = view.to_dict()
    assert vc_dict["component"] == "spatial"

    view = vitessce_config.get_view_by_index(1)
    vc_dict = view.to_dict()
    assert vc_dict["component"] == "scatterplot"

    with pytest.raises(IndexError):
        vitessce_config.get_view_by_index(5)


def test_get_view_by_component(vitessce_config):
    view = vitessce_config.get_first_view_by_type("spatial")
    vc_dict = view.to_dict()
    assert vc_dict["component"] == "spatial"

    view = vitessce_config.get_first_view_by_type("SCATTERPLOT")
    vc_dict = view.to_dict()
    assert vc_dict["component"] == "scatterplot"

    with pytest.raises(ValueError):
        vitessce_config.get_first_view_by_type("TEST")


def test_get_view_by_invalid_type(vitessce_config):
    with pytest.raises(TypeError):
        vitessce_config.get_first_view_by_type(3.5)


def test_remove_view_by_index(vitessce_config):
    removed_view = vitessce_config.remove_view_by_index(0)
    rv_dict = removed_view.to_dict()
    assert rv_dict["component"] == "spatial"
    assert len(vitessce_config.get_views()) == 2

    removed_view = vitessce_config.remove_view_by_index(1)
    rv_dict = removed_view.to_dict()
    assert rv_dict["component"] == "scatterplot"
    assert len(vitessce_config.get_views()) == 1

    with pytest.raises(IndexError):
        vitessce_config.remove_view_by_index(5)


def test_remove_view_by_component(vitessce_config):
    removed_view = vitessce_config.remove_first_view_by_type("spatial")
    rv_dict = removed_view.to_dict()
    assert rv_dict["component"] == "spatial"
    assert len(vitessce_config.get_views()) == 2

    with pytest.raises(ValueError):
        vitessce_config.remove_first_view_by_type("spatial")

    with pytest.raises(ValueError):
        vitessce_config.remove_first_view_by_type("TEST")


def test_remove_view_by_invalid_index(vitessce_config):
    with pytest.raises(TypeError):
        vitessce_config.remove_view_by_index(3.5)



from __future__ import annotations

from typing import TYPE_CHECKING

from collections import defaultdict
import os
from os.path import join
import tempfile
from typing import Callable, Optional, Type, TypeVar, Union
from uuid import uuid4
from pathlib import PurePath, PurePosixPath
import warnings
import zarr

import numpy as np

if TYPE_CHECKING:
    import lamindb as ln
    from spatialdata import SpatialData

from vitessce.file_def_utils import (
    gen_obs_locations_schema,
    gen_obs_segmentations_schema,
    gen_obs_spots_schema,
    gen_obs_points_schema,
    gen_obs_embedding_schema,
    gen_feature_labels_schema,
    gen_obs_feature_matrix_schema,
    gen_obs_labels_schema,
    gen_obs_sets_schema,
    gen_sdata_image_schema,
    gen_sdata_labels_schema,
    gen_sdata_obs_spots_schema,
    gen_sdata_obs_sets_schema,
    gen_sdata_obs_feature_matrix_schema,
)

from .constants import (
    norm_enum,
    ViewType as cm,
    FileType as ft,
    DataType as dt,
)
from .repr import make_repr


def make_unique_filename(file_ext):
    return f"{str(uuid4())}{file_ext}"


def file_path_to_url_path(local_path, prepend_slash=True, path_class=None):
    # force_windows is used in tests
    url_path = str(PurePosixPath(PurePath(local_path) if path_class is None else path_class(local_path)))
    if prepend_slash and not url_path.startswith("/"):
        url_path = f"/{url_path}"
    return url_path


class AbstractWrapper:
    """
    An abstract class that can be extended when
    implementing custom dataset object wrapper classes.
    """

    def __init__(self, **kwargs):
        """
        Abstract constructor to be inherited by dataset wrapper classes.

        :param str out_dir: The path to a local directory used for data processing outputs. By default, uses a temp. directory.
        :param dict request_init: options to be passed along with every fetch request from the browser, like `{ "header": { "Authorization": "Bearer dsfjalsdfa1431" } }`
        """
        self.out_dir = kwargs['out_dir'] if 'out_dir' in kwargs else tempfile.mkdtemp(
        )
        self.routes = []
        self.is_remote = False  # TODO: change to needs_localhost_serving for clarity
        self.is_store = False  # TODO: change to needs_store_registration for clarity
        self.file_def_creators = []
        self.base_dir = None
        self.stores = {}
        self.artifacts = {}
        self._request_init = kwargs['request_init'] if 'request_init' in kwargs else None

    def __repr__(self):
        return self._repr

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        """
        Fill in the file_def_creators array.
        Each function added to this list should take in a base URL and generate a Vitessce file definition.
        If this wrapper is wrapping local data, then create routes and fill in the routes array.
        This method is void, should not return anything.

        :param str dataset_uid: A unique identifier for this dataset.
        :param int obj_i: Within the dataset, the index of this data wrapper object.
        """
        os.makedirs(self._get_out_dir(dataset_uid, obj_i), exist_ok=True)
        self.base_dir = base_dir

    def get_routes(self):
        """
        Obtain the routes that have been created for this wrapper class.

        :returns: A list of server routes.
        :rtype: list[starlette.routing.Route]
        """
        return self.routes

    def register_artifact(self, artifact):
        """
        Register an artifact.

        :param artifact: The artifact object to register.
        :type artifact: lamindb.Artifact
        :returns: The artifact URL.
        :rtype: str
        """
        artifact_url = artifact.path.to_url()
        self.artifacts[artifact_url] = artifact
        return artifact_url

    def get_artifacts(self):
        """
        Obtain the dictionary that maps artifact URLs to artifact objects.

        :returns: A dictionary that maps artifact URLs to Artifact objects.
        :rtype: dict[str, lamindb.Artifact]
        """
        return self.artifacts

    def get_stores(self, base_url):
        """
        Obtain the stores that have been created for this wrapper class.

        :returns: A dictionary that maps file URLs to Zarr Store objects.
        :rtype: dict[str, zarr.Store]
        """
        relative_stores = self.stores
        absolute_stores = {}
        for relative_url, store in relative_stores.items():
            absolute_url = base_url + relative_url
            absolute_stores[absolute_url] = store
        return absolute_stores

    def get_file_defs(self, base_url):
        """
        Obtain the file definitions for this wrapper class.

        :param str base_url: A base URL to prepend to relative URLs.

        :returns: A list of file definitions.
        :rtype: list[dict]
        """
        file_defs_with_base_url = []
        for file_def_creator in self.file_def_creators:
            file_def = file_def_creator(base_url)
            if file_def is not None:
                file_defs_with_base_url.append(file_def)
        return file_defs_with_base_url

    def get_out_dir_route(self, dataset_uid, obj_i):
        """
        Obtain the Mount for the `out_dir`

        :param str dataset_uid: A dataset unique identifier for the Mount
        :param str obj_i: A index of the current vitessce.wrappers.AbstractWrapper among all other wrappers in the view config

        :returns: A starlette Mount of the the `out_dir`
        :rtype: list[starlette.routing.Mount]
        """
        if not self.is_remote:
            out_dir = self._get_out_dir(dataset_uid, obj_i)
            # TODO: Move imports back to top when this is factored out.
            from starlette.staticfiles import StaticFiles
            from starlette.routing import Mount
            return [Mount(self._get_route_str(dataset_uid, obj_i),
                          app=StaticFiles(directory=out_dir, html=False))]
        return []

    def get_local_file_url(self, base_url, dataset_uid, obj_i, local_file_path, local_file_uid):
        if not self.is_remote and self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(local_file_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid, obj_i, local_file_uid)

    def get_local_dir_url(self, base_url, dataset_uid, obj_i, local_dir_path, local_dir_uid):
        # Logic for files and directories is the same for this function.
        return self.get_local_file_url(base_url, dataset_uid, obj_i, local_dir_path, local_dir_uid)

    def register_zarr_store(self, dataset_uid, obj_i, store_or_local_dir_path, local_dir_uid):
        if not self.is_remote and self.is_store:
            # Set up `store` and `local_dir_path` variables.
            if isinstance(store_or_local_dir_path, str):
                # TODO: use zarr.FSStore if fsspec is installed?
                store = zarr.DirectoryStore(store_or_local_dir_path)
                local_dir_path = store_or_local_dir_path
            else:
                # TODO: check that store_or_local_dir_path is a zarr.Store or StoreLike?
                store = store_or_local_dir_path
                # A store instance was passed directly, so there is no local directory path.
                # Instead we just make one up using _get_route_str but it could be any string.
                local_dir_path = self._get_route_str(dataset_uid, obj_i, local_dir_uid)

            # Register the store on the same route path
            # that will be used for the "url" field in the file definition.
            if self.base_dir is None:
                route_path = self._get_route_str(dataset_uid, obj_i, local_dir_uid)
            else:
                route_path = file_path_to_url_path(local_dir_path)
                local_dir_path = join(self.base_dir, local_dir_path)

            self.stores[route_path] = store

    def get_local_dir_route(self, dataset_uid, obj_i, local_dir_path, local_dir_uid):
        """
        Obtain the Mount for some local directory

        :param str dataset_uid: A dataset unique identifier for the Mount
        :param str obj_i: A index of the current vitessce.wrappers.AbstractWrapper among all other wrappers in the view config
        :param str local_dir_path: The path to the local directory to serve.
        :param str local_dir_uid: The UID to include as the route path suffix.

        :returns: A starlette Mount of the the `local_dir_path`
        :rtype: list[starlette.routing.Mount]
        """
        if not self.is_remote:
            if self.base_dir is None:
                route_path = self._get_route_str(dataset_uid, obj_i, local_dir_uid)
            else:
                route_path = file_path_to_url_path(local_dir_path)
                local_dir_path = join(self.base_dir, local_dir_path)
            # TODO: Move imports back to top when this is factored out.
            from starlette.staticfiles import StaticFiles
            from starlette.routing import Mount
            return [Mount(route_path,
                          app=StaticFiles(directory=local_dir_path, html=False))]
        return []

    def get_local_file_route(self, dataset_uid, obj_i, local_file_path, local_file_uid):
        if not self.is_remote:
            from .routes import range_repsonse, FileRoute

            if self.base_dir is None:
                route_path = self._get_route_str(dataset_uid, obj_i, local_file_uid)
            else:
                route_path = file_path_to_url_path(local_file_path)
                local_file_path = join(self.base_dir, local_file_path)

            return [
                FileRoute(route_path, lambda req: range_repsonse(req, local_file_path), local_file_path),
            ]
        return []

    def _get_url(self, base_url, dataset_uid, obj_i, *args):
        return base_url + self._get_route_str(dataset_uid, obj_i, *args)

    def _get_url_simple(self, base_url, suffix):
        return base_url + "/" + suffix

    def _get_route_str(self, dataset_uid, obj_i, *args):
        return "/" + "/".join(map(str, [dataset_uid, obj_i, *args]))

    def _get_out_dir(self, dataset_uid, obj_i, *args):
        return join(self.out_dir, dataset_uid, str(obj_i), *args)

    def auto_view_config(self, vc):
        """
        Auto view configuration is intended to be used internally by the `VitessceConfig.from_object` method.
        Each subclass of `AbstractWrapper` may implement this method which takes in a `VitessceConfig` instance
        and modifies it by adding datasets, visualization components, and view coordinations.
        Implementations of this method may create an opinionated view config based on inferred use cases.

        :param vc: The view config instance.
        :type vc: VitessceConfig
        """
        raise NotImplementedError(
            "Auto view configuration has not yet been implemented for this data object wrapper class.")


class MultiImageWrapper(AbstractWrapper):
    """
    Wrap multiple imaging datasets by creating an instance of the ``MultiImageWrapper`` class.

    :param list image_wrappers: A list of imaging wrapper classes (only :class:`~vitessce.wrappers.OmeTiffWrapper` supported now)
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, image_wrappers, use_physical_size_scaling=False, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        self.image_wrappers = image_wrappers
        self.use_physical_size_scaling = use_physical_size_scaling

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        for image in self.image_wrappers:
            image.convert_and_save(dataset_uid, obj_i, base_dir=base_dir)
        file_def_creator = self.make_raster_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_raster_routes()
        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_raster_routes(self):
        obj_routes = []
        for num, image in enumerate(self.image_wrappers):
            obj_routes = obj_routes + image.get_routes()
        return obj_routes

    def make_raster_file_def_creator(self, dataset_uid, obj_i):

        def raster_file_def_creator(base_url):
            raster_json = {
                "schemaVersion": "0.0.2",
                "usePhysicalSizeScaling": self.use_physical_size_scaling,
                "images": [],
                "renderLayers": []
            }
            for image in self.image_wrappers:
                image_json = image.make_image_def(dataset_uid, obj_i, base_url)
                raster_json['images'].append(image_json)
                raster_json['renderLayers'].append(image.name)

            return {
                "fileType": ft.RASTER_JSON.value,
                "options": raster_json
            }

        return raster_file_def_creator


class OmeTiffWrapper(AbstractWrapper):

    """
    Wrap an OME-TIFF File by creating an instance of the ``OmeTiffWrapper`` class.

    :param str img_path: A local filepath to an OME-TIFF file.
    :param str offsets_path: A local filepath to an offsets.json file.
    :param str img_url: A remote URL of an OME-TIFF file.
    :param str offsets_url: A remote URL of an offsets.json file.
    :param str name: The display name for this OME-TIFF within Vitessce.
    :param list[number] transformation_matrix: A column-major ordered matrix for transforming this image (see http://www.opengl-tutorial.org/beginners-tutorials/tutorial-3-matrices/#homogeneous-coordinates for more information).
    :param bool is_bitmask: Whether or not this image is a bitmask.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, offsets_path=None, img_url=None, offsets_url=None, name="", transformation_matrix=None, is_bitmask=False,
                 **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        self.name = name
        self._img_path = img_path
        self._img_url = img_url
        self._offsets_url = offsets_url
        self._transformation_matrix = transformation_matrix
        self.is_remote = img_url is not None
        self.is_bitmask = is_bitmask
        self.local_img_uid = make_unique_filename(".ome.tif")
        self.local_offsets_uid = make_unique_filename(".offsets.json")
        if img_url is not None and (img_path is not None or offsets_path is not None):
            raise ValueError(
                "Did not expect img_path or offsets_path to be provided with img_url")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_raster_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_raster_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_raster_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            # TODO: Move imports back to top when this is factored out.
            from .routes import range_repsonse, JsonRoute, FileRoute
            from generate_tiff_offsets import get_offsets
            from starlette.responses import UJSONResponse

            offsets = get_offsets(self._img_path)

            async def response_func(req):
                return UJSONResponse(offsets)
            if self.base_dir is None:
                local_img_path = self._img_path
                local_img_route_path = self._get_route_str(dataset_uid, obj_i, self.local_img_uid)
                local_offsets_route_path = self._get_route_str(dataset_uid, obj_i, self.local_offsets_uid)
            else:
                local_img_path = join(self.base_dir, self._img_path)
                local_img_route_path = file_path_to_url_path(self._img_path)
                # Do not include offsets in base_dir mode.
                local_offsets_route_path = None

            routes = [
                FileRoute(local_img_route_path, lambda req: range_repsonse(req, local_img_path), local_img_path),
            ]
            if local_offsets_route_path is not None:
                # Do not include offsets in base_dir mode.
                routes.append(JsonRoute(local_offsets_route_path, response_func, offsets))

            return routes

    def make_image_def(self, dataset_uid, obj_i, base_url):
        img_url = self.get_img_url(base_url, dataset_uid, obj_i)
        offsets_url = self.get_offsets_url(base_url, dataset_uid, obj_i)
        return self.create_image_json(img_url, offsets_url)

    def make_raster_file_def_creator(self, dataset_uid, obj_i):
        def raster_file_def_creator(base_url):
            raster_json = {
                "schemaVersion": "0.0.2",
                "images": [self.make_image_def(dataset_uid, obj_i, base_url)],
            }

            return {
                "fileType": ft.RASTER_JSON.value,
                "options": raster_json
            }
        return raster_file_def_creator

    def create_image_json(self, img_url, offsets_url=None):
        metadata = {}
        image = {
            "name": self.name,
            "type": "ome-tiff",
            "url": img_url,
        }
        if offsets_url is not None and self.base_dir is None:
            # Do not include offsets in base_dir mode.
            metadata["omeTiffOffsetsUrl"] = offsets_url
        if self._transformation_matrix is not None:
            metadata["transform"] = {
                "matrix": self._transformation_matrix
            }
        metadata["isBitmask"] = self.is_bitmask
        # Only attach metadata if there is some - otherwise schema validation fails.
        if len(metadata.keys()) > 0:
            image["metadata"] = metadata
        return image

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        if self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(self._img_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid,
                             obj_i, self.local_img_uid)

    def get_offsets_url(self, base_url="", dataset_uid="", obj_i=""):
        if self._offsets_url is not None or self.is_remote:
            return self._offsets_url
        offsets_url = self._get_url(
            base_url, dataset_uid, obj_i, self.local_offsets_uid)
        return offsets_url


class ImageOmeTiffWrapper(AbstractWrapper):

    """
    Wrap an OME-TIFF File by creating an instance of the ``ImageOmeTiffWrapper`` class. Intended to be used with the spatialBeta and layerControllerBeta views.

    :param str img_path: A local filepath to an OME-TIFF file.
    :param str offsets_path: A local filepath to an offsets.json file.
    :param str img_url: A remote URL of an OME-TIFF file.
    :param str offsets_url: A remote URL of an offsets.json file.
    :param list coordinate_transformations: A column-major ordered matrix for transforming this image (see http://www.opengl-tutorial.org/beginners-tutorials/tutorial-3-matrices/#homogeneous-coordinates for more information).
    :param dict coordination_values: Optional coordinationValues to be passed in the file definition.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, img_url=None, img_artifact=None, offsets_path=None, offsets_url=None, offsets_artifact=None, coordinate_transformations=None, coordination_values=None, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        num_inputs = sum([1 for x in [img_path, img_url, img_artifact] if x is not None])
        if num_inputs != 1:
            raise ValueError(
                "Expected one of img_path, img_url, or img_artifact to be provided")

        num_inputs = sum([1 for x in [offsets_path, offsets_url, offsets_artifact] if x is not None])
        if num_inputs > 1:
            raise ValueError(
                "Expected zero or one of offsets_path, offsets_url, or offsets_artifact to be provided")

        self._img_path = img_path
        self._img_url = img_url
        self._img_artifact = img_artifact
        self._offsets_path = offsets_path
        self._offsets_url = offsets_url
        self._offsets_artifact = offsets_artifact
        self._coordinate_transformations = coordinate_transformations
        self._coordination_values = coordination_values
        self.is_remote = img_url is not None or img_artifact is not None
        self.local_img_uid = make_unique_filename(".ome.tif")
        self.local_offsets_uid = make_unique_filename(".offsets.json")

        if img_artifact is not None:
            self._img_url = self.register_artifact(img_artifact)

        if offsets_artifact is not None:
            self._offsets_url = self.register_artifact(offsets_artifact)

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_raster_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_raster_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_raster_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            # TODO: Move imports back to top when this is factored out.
            from .routes import range_repsonse, JsonRoute, FileRoute
            from generate_tiff_offsets import get_offsets
            from starlette.responses import UJSONResponse

            offsets = get_offsets(self._img_path)

            async def response_func(req):
                return UJSONResponse(offsets)
            if self.base_dir is None:
                local_img_path = self._img_path
                local_img_route_path = self._get_route_str(dataset_uid, obj_i, self.local_img_uid)
                local_offsets_route_path = self._get_route_str(dataset_uid, obj_i, self.local_offsets_uid)
            else:
                local_img_path = join(self.base_dir, self._img_path)
                local_img_route_path = file_path_to_url_path(self._img_path)
                # Do not include offsets in base_dir mode.
                local_offsets_route_path = None

            routes = [
                FileRoute(local_img_route_path, lambda req: range_repsonse(req, local_img_path), local_img_path),
            ]
            if local_offsets_route_path is not None:
                # Do not include offsets in base_dir mode.
                routes.append(JsonRoute(local_offsets_route_path, response_func, offsets))

            return routes

    def make_raster_file_def_creator(self, dataset_uid, obj_i):
        def raster_file_def_creator(base_url):
            options = {}
            if self._coordinate_transformations is not None:
                options["coordinateTransformations"] = self._coordinate_transformations

            offsets_url = self.get_offsets_url(base_url, dataset_uid, obj_i)
            if offsets_url is not None and self.base_dir is None:
                # Do not include offsets in base_dir mode.
                options["offsetsUrl"] = offsets_url

            file_def = {
                "fileType": "image.ome-tiff",
                "url": self.get_img_url(base_url, dataset_uid, obj_i),
            }
            if len(options.keys()) > 0:
                file_def["options"] = options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def
        return raster_file_def_creator

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        if self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(self._img_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid,
                             obj_i, self.local_img_uid)

    def get_offsets_url(self, base_url="", dataset_uid="", obj_i=""):
        if self._offsets_url is not None or self.is_remote:
            return self._offsets_url
        offsets_url = self._get_url(
            base_url, dataset_uid, obj_i, self.local_offsets_uid)
        return offsets_url


class ObsSegmentationsOmeTiffWrapper(AbstractWrapper):

    """
    Wrap an OME-TIFF File by creating an instance of the ``ObsSegmentationsOmeTiffWrapper`` class. Intended to be used with the spatialBeta and layerControllerBeta views.

    :param str img_path: A local filepath to an OME-TIFF file.
    :param str img_url: A remote URL of an OME-TIFF file.
    :param img_artifact: A lamindb Artifact corresponding to the image.
    :type img_artifact: lamindb.Artifact
    :param str offsets_path: A local filepath to an offsets.json file.
    :param str offsets_url: A remote URL of an offsets.json file.
    :param offsets_artifact: A lamindb Artifact corresponding to the offsets JSON.
    :type offsets_artifact: lamindb.Artifact
    :param list coordinate_transformations: A column-major ordered matrix for transforming this image (see http://www.opengl-tutorial.org/beginners-tutorials/tutorial-3-matrices/#homogeneous-coordinates for more information).
    :param bool obs_types_from_channel_names: Whether to use the channel names to determine the obs types. Optional.
    :param dict coordination_values: Optional coordinationValues to be passed in the file definition.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, img_url=None, img_artifact=None, offsets_path=None, offsets_url=None, offsets_artifact=None, coordinate_transformations=None, obs_types_from_channel_names=None, coordination_values=None, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        num_inputs = sum([1 for x in [img_path, img_url, img_artifact] if x is not None])
        if num_inputs != 1:
            raise ValueError(
                "Expected one of img_path, img_url, or img_artifact to be provided")

        num_inputs = sum([1 for x in [offsets_path, offsets_url, offsets_artifact] if x is not None])
        if num_inputs > 1:
            raise ValueError(
                "Expected zero or one of offsets_path, offsets_url, or offsets_artifact to be provided")

        self._img_path = img_path
        self._img_url = img_url
        self._img_artifact = img_artifact
        self._offsets_path = offsets_path
        self._offsets_url = offsets_url
        self._offsets_artifact = offsets_artifact
        self._coordinate_transformations = coordinate_transformations
        self._obs_types_from_channel_names = obs_types_from_channel_names
        self._coordination_values = coordination_values
        self.is_remote = img_url is not None or img_artifact is not None
        self.local_img_uid = make_unique_filename(".ome.tif")
        self.local_offsets_uid = make_unique_filename(".offsets.json")

        if img_artifact is not None:
            self._img_url = self.register_artifact(img_artifact)

        if offsets_artifact is not None:
            self._offsets_url = self.register_artifact(offsets_artifact)

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_raster_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_raster_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_raster_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            # TODO: Move imports back to top when this is factored out.
            from .routes import range_repsonse, JsonRoute, FileRoute
            from generate_tiff_offsets import get_offsets
            from starlette.responses import UJSONResponse

            offsets = get_offsets(self._img_path)

            async def response_func(req):
                return UJSONResponse(offsets)
            if self.base_dir is None:
                local_img_path = self._img_path
                local_img_route_path = self._get_route_str(dataset_uid, obj_i, self.local_img_uid)
                local_offsets_route_path = self._get_route_str(dataset_uid, obj_i, self.local_offsets_uid)
            else:
                local_img_path = join(self.base_dir, self._img_path)
                local_img_route_path = file_path_to_url_path(self._img_path)
                # Do not include offsets in base_dir mode.
                local_offsets_route_path = None

            routes = [
                FileRoute(local_img_route_path, lambda req: range_repsonse(req, local_img_path), local_img_path),
            ]
            if local_offsets_route_path is not None:
                # Do not include offsets in base_dir mode.
                routes.append(JsonRoute(local_offsets_route_path, response_func, offsets))

            return routes

    def make_raster_file_def_creator(self, dataset_uid, obj_i):
        def raster_file_def_creator(base_url):
            options = {}
            if self._coordinate_transformations is not None:
                options["coordinateTransformations"] = self._coordinate_transformations

            if self._obs_types_from_channel_names is not None:
                options["obsTypesFromChannelNames"] = self._obs_types_from_channel_names

            offsets_url = self.get_offsets_url(base_url, dataset_uid, obj_i)
            if offsets_url is not None and self.base_dir is None:
                # Do not include offsets in base_dir mode.
                options["offsetsUrl"] = offsets_url

            file_def = {
                "fileType": "obsSegmentations.ome-tiff",
                "url": self.get_img_url(base_url, dataset_uid, obj_i),
            }
            if len(options.keys()) > 0:
                file_def["options"] = options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def
        return raster_file_def_creator

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        if self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(self._img_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid,
                             obj_i, self.local_img_uid)

    def get_offsets_url(self, base_url="", dataset_uid="", obj_i=""):
        if self._offsets_url is not None or self.is_remote:
            return self._offsets_url
        offsets_url = self._get_url(
            base_url, dataset_uid, obj_i, self.local_offsets_uid)
        return offsets_url


class CsvWrapper(AbstractWrapper):

    """
    Wrap a CSV file by creating an instance of the ``CsvWrapper`` class.

    :param str data_type: The data type of the information contained in the file.
    :param str csv_path: A local filepath to a CSV file.
    :param str csv_url: A remote URL of a CSV file.
    :param dict options: The file options.
    :param dict coordination_values: The coordination values.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, csv_path=None, csv_url=None, data_type=None, options=None, coordination_values=None,
                 **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        self._csv_path = csv_path
        self._csv_url = csv_url
        self._data_type = norm_enum(data_type, dt)
        self._options = options
        self._coordination_values = coordination_values
        self.is_remote = csv_url is not None
        self.local_csv_uid = make_unique_filename(".csv")
        if data_type is None:
            raise ValueError("Expected data_type to be provided")
        if csv_url is not None and csv_path is not None:
            raise ValueError(
                "Did not expect csv_url to be provided with csv_path")
        if csv_url is None and csv_path is None:
            raise ValueError(
                "Expected csv_url or csv_path to be provided")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_csv_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_csv_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_csv_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            # TODO: Move imports back to top when this is factored out.
            from .routes import FileRoute
            from starlette.responses import FileResponse

            if self.base_dir is not None:
                local_csv_path = join(self.base_dir, self._csv_path)
                local_csv_route_path = file_path_to_url_path(self._csv_path)
            else:
                local_csv_path = self._csv_path
                local_csv_route_path = self._get_route_str(dataset_uid, obj_i, self.local_csv_uid)

            async def response_func(req):
                return FileResponse(local_csv_path, filename=os.path.basename(self._csv_path))
            routes = [
                FileRoute(local_csv_route_path, response_func, local_csv_path),
            ]
            return routes

    def make_csv_file_def_creator(self, dataset_uid, obj_i):
        def csv_file_def_creator(base_url):
            file_def = {
                "fileType": f"{self._data_type}.csv",
                "url": self.get_csv_url(base_url, dataset_uid, obj_i),
            }
            if self._options is not None:
                file_def["options"] = self._options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def
        return csv_file_def_creator

    def get_csv_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._csv_url
        if self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(self._csv_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid,
                             obj_i, self.local_csv_uid)


class JsonWrapper(AbstractWrapper):

    """
    Wrap a JSON file by creating an instance of the ``JsonWrapper`` class.

    :param str data_type: The data type of the information contained in the file.
    :param str json_path: A local filepath JSON a JSON file.
    :param str json_url: A remote URL of a CSV file.
    :param dict options: The file options.
    :param dict coordination_values: The coordination values.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, json_path=None, json_url=None, data_type=None, options=None, coordination_values=None,
                 **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        self._json_path = json_path
        self._json_url = json_url
        self._data_type = norm_enum(data_type, dt)
        self._options = options
        self._coordination_values = coordination_values
        self.is_remote = json_url is not None
        self.local_json_uid = make_unique_filename(".json")
        if data_type is None:
            raise ValueError("Expected data_type to be provided")
        if json_url is not None and json_path is not None:
            raise ValueError(
                "Did not expect json_url to be provided with json_path")
        if json_url is None and json_path is None:
            raise ValueError(
                "Expected json_url or json_path to be provided")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_json_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_json_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_json_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            # TODO: Move imports back to top when this is factored out.
            from .routes import FileRoute
            from starlette.responses import FileResponse

            if self.base_dir is not None:
                local_json_path = join(self.base_dir, self._json_path)
                local_json_route_path = file_path_to_url_path(self._json_path)
            else:
                local_json_path = self._json_path
                local_json_route_path = self._get_route_str(dataset_uid, obj_i, self.local_json_uid)

            async def response_func(req):
                return FileResponse(local_json_path, filename=os.path.basename(self._json_path))
            routes = [
                FileRoute(local_json_route_path, response_func, local_json_path),
            ]
            return routes

    def make_json_file_def_creator(self, dataset_uid, obj_i):
        def json_file_def_creator(base_url):
            file_def = {
                "fileType": f"{self._data_type}.json",
                "url": self.get_json_url(base_url, dataset_uid, obj_i),
            }
            if self._options is not None:
                file_def["options"] = self._options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def
        return json_file_def_creator

    def get_json_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._json_url
        if self.base_dir is not None:
            return self._get_url_simple(base_url, file_path_to_url_path(self._json_path, prepend_slash=False))
        return self._get_url(base_url, dataset_uid,
                             obj_i, self.local_json_uid)


class OmeZarrWrapper(AbstractWrapper):

    """
    Wrap an OME-NGFF Zarr store by creating an instance of the ``OmeZarrWrapper`` class.

    :param str img_path: A local filepath to an OME-NGFF Zarr store.
    :param str img_url: A remote URL of an OME-NGFF Zarr store.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, img_url=None, name="", is_bitmask=False, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        if img_url is not None and img_path is not None:
            raise ValueError(
                "Did not expect img_path to be provided with img_url")
        if img_url is None and img_path is None:
            raise ValueError(
                "Expected either img_url or img_path to be provided")
        self._img_path = img_path
        self._img_url = img_url
        self.name = name
        self.is_bitmask = is_bitmask
        if self._img_path is not None:
            self.is_remote = False
        else:
            self.is_remote = True
        self.local_dir_uid = make_unique_filename(".ome.zarr")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_image_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_image_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_image_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            return self.get_local_dir_route(dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        return self.get_local_dir_url(base_url, dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def make_image_file_def_creator(self, dataset_uid, obj_i):
        def image_file_def_creator(base_url):
            return {
                "fileType": "image.ome-zarr",
                "url": self.get_img_url(base_url, dataset_uid, obj_i)
            }
        return image_file_def_creator

    # The following two functions will be used when OmeZarrWrapper
    # is used within MultiImageWrapper.
    def make_image_def(self, dataset_uid, obj_i, base_url):
        img_url = self.get_img_url(base_url, dataset_uid, obj_i)
        return self.create_image_json(img_url)

    def create_image_json(self, img_url):
        metadata = {}
        image = {
            "name": self.name,
            "type": "ome-zarr",
            "url": img_url,
        }
        if self.is_bitmask:
            metadata["isBitmask"] = self.is_bitmask
        # Only attach metadata if there is some - otherwise schema validation fails.
        if len(metadata.keys()) > 0:
            image["metadata"] = metadata
        return image


class ImageOmeZarrWrapper(AbstractWrapper):

    """
    Wrap an OME-NGFF Zarr store by creating an instance of the ``ImageOmeZarrWrapper`` class. Intended to be used with the spatialBeta and layerControllerBeta views.

    :param str img_path: A local filepath to an OME-NGFF Zarr store.
    :param str img_url: A remote URL of an OME-NGFF Zarr store.
    :param img_artifact: A lamindb Artifact corresponding to the image.
    :type img_artifact: lamindb.Artifact
    :param list coordinate_transformations: A column-major ordered matrix for transforming this image (see http://www.opengl-tutorial.org/beginners-tutorials/tutorial-3-matrices/#homogeneous-coordinates for more information).
    :param dict coordination_values: Optional coordinationValues to be passed in the file definition.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, img_url=None, img_artifact=None, coordinate_transformations=None, coordination_values=None, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())

        num_inputs = sum([1 for x in [img_path, img_url, img_artifact] if x is not None])
        if num_inputs != 1:
            raise ValueError(
                "Expected one of img_path, img_url, or img_artifact to be provided")

        self._img_path = img_path
        self._img_url = img_url
        self._img_artifact = img_artifact
        self._coordinate_transformations = coordinate_transformations
        self._coordination_values = coordination_values
        if self._img_path is not None:
            self.is_remote = False
        else:
            self.is_remote = True

        if self._img_artifact is not None:
            # To serve as a placeholder in the config JSON URL field
            self._img_url = self.register_artifact(img_artifact)

        self.local_dir_uid = make_unique_filename(".ome.zarr")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_image_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_image_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_image_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            return self.get_local_dir_route(dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        return self.get_local_dir_url(base_url, dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def make_image_file_def_creator(self, dataset_uid, obj_i):
        def image_file_def_creator(base_url):
            options = {}
            if self._coordinate_transformations is not None:
                options["coordinateTransformations"] = self._coordinate_transformations

            file_def = {
                "fileType": "image.ome-zarr",
                "url": self.get_img_url(base_url, dataset_uid, obj_i)
            }

            if len(options.keys()) > 0:
                file_def["options"] = options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def

        return image_file_def_creator


class ObsSegmentationsOmeZarrWrapper(AbstractWrapper):

    """
    Wrap an OME-NGFF Zarr store by creating an instance of the ``ObsSegmentationsOmeZarrWrapper`` class. Intended to be used with the spatialBeta and layerControllerBeta views.

    :param str img_path: A local filepath to an OME-NGFF Zarr store.
    :param str img_url: A remote URL of an OME-NGFF Zarr store.
    :param img_artifact: A lamindb Artifact corresponding to the image.
    :type img_artifact: lamindb.Artifact
    :param list coordinate_transformations: A column-major ordered matrix for transforming this image (see http://www.opengl-tutorial.org/beginners-tutorials/tutorial-3-matrices/#homogeneous-coordinates for more information).
    :param dict coordination_values: Optional coordinationValues to be passed in the file definition.
    :param bool obs_types_from_channel_names: Whether to use the channel names to determine the obs types. Optional.
    :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
    """

    def __init__(self, img_path=None, img_url=None, img_artifact=None, coordinate_transformations=None, coordination_values=None, obs_types_from_channel_names=None, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())

        num_inputs = sum([1 for x in [img_path, img_url, img_artifact] if x is not None])
        if num_inputs != 1:
            raise ValueError(
                "Expected one of img_path, img_url, or img_artifact to be provided")
        self._img_path = img_path
        self._img_url = img_url
        self._img_artifact = img_artifact
        self._coordinate_transformations = coordinate_transformations
        self._obs_types_from_channel_names = obs_types_from_channel_names
        self._coordination_values = coordination_values
        if self._img_path is not None:
            self.is_remote = False
        else:
            self.is_remote = True

        if self._img_artifact is not None:
            # To serve as a placeholder in the config JSON URL field
            self._img_url = self.register_artifact(img_artifact)

        self.local_dir_uid = make_unique_filename(".ome.zarr")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_image_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_image_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_image_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            return self.get_local_dir_route(dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def get_img_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._img_url
        return self.get_local_dir_url(base_url, dataset_uid, obj_i, self._img_path, self.local_dir_uid)

    def make_image_file_def_creator(self, dataset_uid, obj_i):
        def image_file_def_creator(base_url):
            options = {}
            if self._coordinate_transformations is not None:
                options["coordinateTransformations"] = self._coordinate_transformations

            if self._obs_types_from_channel_names is not None:
                options["obsTypesFromChannelNames"] = self._obs_types_from_channel_names

            file_def = {
                "fileType": "obsSegmentations.ome-zarr",
                "url": self.get_img_url(base_url, dataset_uid, obj_i)
            }

            if len(options.keys()) > 0:
                file_def["options"] = options
            if self._coordination_values is not None:
                file_def["coordinationValues"] = self._coordination_values
            return file_def

        return image_file_def_creator


def raise_error_if_zero_or_more_than_one(inputs):
    num_inputs = sum([1 for x in inputs if x is not None])
    if num_inputs > 1:
        raise ValueError(
            "Expected only one type of data input parameter to be provided (_url, _path, _store, etc.), but received more than one."
        )
    if num_inputs == 0:
        raise ValueError(
            "Expected one type of data input parameter to be provided (_url, _path, _store, etc.), but received none."
        )
    return True


def raise_error_if_any(inputs):
    num_inputs = sum([1 for x in inputs if x is not None])
    if num_inputs > 0:
        raise ValueError(
            "Did not expect any of these parameters to be provided, but received one or more: " + str(inputs)
        )
    return True


def raise_error_if_more_than_one(inputs):
    num_inputs = sum([1 for x in inputs if x is not None])
    if num_inputs > 1:
        raise ValueError(
            "Expected only one of these parameters to be provided, but received more than one: " + str(inputs)
        )
    return True


class AnnDataWrapper(AbstractWrapper):
    def __init__(self, adata_path=None, adata_url=None, adata_store=None, adata_artifact=None, ref_path=None, ref_url=None, ref_artifact=None, obs_feature_matrix_path=None, feature_filter_path=None, initial_feature_filter_path=None, obs_set_paths=None, obs_set_names=None, obs_locations_path=None, obs_segmentations_path=None, obs_embedding_paths=None, obs_embedding_names=None, obs_embedding_dims=None, obs_spots_path=None, obs_points_path=None, feature_labels_path=None, obs_labels_path=None, convert_to_dense=True, coordination_values=None, obs_labels_paths=None, obs_labels_names=None, **kwargs):
        """
        Wrap an AnnData object by creating an instance of the ``AnnDataWrapper`` class.

        :param str adata_path: A path to an AnnData object written to a Zarr store containing single-cell experiment data.
        :param str adata_url: A remote url pointing to a zarr-backed AnnData store.
        :param adata_store: A path to pass to zarr.DirectoryStore, or an existing store instance.
        :type adata_store: str or zarr.Storage
        :param adata_artifact: A lamindb Artifact corresponding to the AnnData object.
        :type adata_artifact: lamindb.Artifact
        :param str obs_feature_matrix_path: Location of the expression (cell x gene) matrix, like `X` or `obsm/highly_variable_genes_subset`
        :param str feature_filter_path: A string like `var/highly_variable` used in conjunction with `obs_feature_matrix_path` if obs_feature_matrix_path points to a subset of `X` of the full `var` list.
        :param str initial_feature_filter_path: A string like `var/highly_variable` used in conjunction with `obs_feature_matrix_path` if obs_feature_matrix_path points to a subset of `X` of the full `var` list.
        :param list[str] obs_set_paths: Column names like `['obs/louvain', 'obs/cellType']` for showing cell sets
        :param list[str] obs_set_names: Names to display in place of those in `obs_set_paths`, like `['Louvain', 'Cell Type']`
        :param str obs_locations_path: Column name in `obsm` that contains centroid coordinates for displaying centroids in the spatial viewer
        :param str obs_segmentations_path: Column name in `obsm` that contains polygonal coordinates for displaying outlines in the spatial viewer
        :param list[str] obs_embedding_paths: Column names like `['obsm/X_umap', 'obsm/X_pca']` for showing scatterplots
        :param list[str] obs_embedding_names: Overriding names like `['UMAP', 'PCA']` for displaying above scatterplots
        :param list[str] obs_embedding_dims: Dimensions along which to get data for the scatterplot, like `[[0, 1], [4, 5]]` where `[0, 1]` is just the normal x and y but `[4, 5]` could be comparing the third and fourth principal components, for example.
        :param str obs_spots_path: Column name in `obsm` that contains centroid coordinates for displaying spots in the spatial viewer
        :param str obs_points_path: Column name in `obsm` that contains centroid coordinates for displaying points in the spatial viewer
        :param str feature_labels_path: The name of a column containing feature labels (e.g., alternate gene symbols), instead of the default index in `var` of the AnnData store.
        :param str obs_labels_path: (DEPRECATED) The name of a column containing observation labels (e.g., alternate cell IDs), instead of the default index in `obs` of the AnnData store. Use `obs_labels_paths` and `obs_labels_names` instead. This arg will be removed in a future release.
        :param list[str] obs_labels_paths: The names of columns containing observation labels (e.g., alternate cell IDs), instead of the default index in `obs` of the AnnData store.
        :param list[str] obs_labels_names: The optional display names of columns containing observation labels (e.g., alternate cell IDs), instead of the default index in `obs` of the AnnData store.
        :param bool convert_to_dense: Whether or not to convert `X` to dense the zarr store (dense is faster but takes more disk space).
        :param coordination_values: Coordination values for the file definition.
        :type coordination_values: dict or None
        :param \\*\\*kwargs: Keyword arguments inherited from :class:`~vitessce.wrappers.AbstractWrapper`
        """
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        self._adata_path = adata_path
        self._adata_url = adata_url
        self._adata_store = adata_store
        self._adata_artifact = adata_artifact

        # For reference spec JSON with .h5ad files
        self._ref_path = ref_path
        self._ref_url = ref_url
        self._ref_artifact = ref_artifact

        if ref_path is not None or ref_url is not None or ref_artifact is not None:
            self.is_h5ad = True
        else:
            self.is_h5ad = False

        if adata_store is not None and (ref_path is not None or ref_url is not None or ref_artifact is not None):
            raise ValueError(
                "Did not expect reference JSON to be provided with adata_store")

        raise_error_if_zero_or_more_than_one([adata_path, adata_url, adata_store, adata_artifact])

        if adata_path is not None:
            self.is_remote = False
            self.is_store = False
            self.zarr_folder = 'anndata.zarr'
        elif adata_url is not None or adata_artifact is not None:
            self.is_remote = True
            self.is_store = False
            self.zarr_folder = None

            # Store artifacts on AbstractWrapper.artifacts for downstream access,
            # e.g. in lamindb.save_vitessce_config
            if adata_artifact is not None:
                self._adata_url = self.register_artifact(adata_artifact)
            if ref_artifact is not None:
                self._ref_url = self.register_artifact(ref_artifact)
        else:
            # Store case
            self.is_remote = False
            self.is_store = True
            self.zarr_folder = None

        self.local_dir_uid = make_unique_filename(".adata.zarr")
        self.local_file_uid = make_unique_filename(".h5ad")
        self.local_ref_uid = make_unique_filename(".ref.json")

        self._expression_matrix = obs_feature_matrix_path
        self._obs_set_names = obs_set_names
        self._mappings_obsm_names = obs_embedding_names
        self._gene_var_filter = feature_filter_path
        self._matrix_gene_var_filter = initial_feature_filter_path
        self._obs_set_elems = obs_set_paths
        self._spatial_centroid_obsm = obs_locations_path
        self._spatial_polygon_obsm = obs_segmentations_path
        self._mappings_obsm = obs_embedding_paths
        self._mappings_obsm_dims = obs_embedding_dims
        self._spatial_spots_obsm = obs_spots_path
        self._spatial_points_obsm = obs_points_path
        self._feature_labels = feature_labels_path
        # Support legacy provision of single obs labels path
        if (obs_labels_path is not None):
            warnings.warn("`obs_labels_path` will be deprecated in a future release.", DeprecationWarning)
            self._obs_labels_elems = [obs_labels_path]
            self._obs_labels_names = [obs_labels_path.split('/')[-1]]
        else:
            self._obs_labels_elems = obs_labels_paths
            self._obs_labels_names = obs_labels_names
        self._convert_to_dense = convert_to_dense
        self._coordination_values = coordination_values

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        elif self.is_store:
            self.register_zarr_store(dataset_uid, obj_i, self._adata_store, self.local_dir_uid)
            return []
        else:
            if self.is_h5ad:
                return [
                    *self.get_local_file_route(dataset_uid, obj_i, self._adata_path, self.local_file_uid),
                    *self.get_local_file_route(dataset_uid, obj_i, self._ref_path, self.local_ref_uid)
                ]
            else:
                return self.get_local_dir_route(dataset_uid, obj_i, self._adata_path, self.local_dir_uid)

    def get_zarr_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._adata_url
        else:
            return self.get_local_dir_url(base_url, dataset_uid, obj_i, self._adata_path, self.local_dir_uid)

    def get_h5ad_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._adata_url
        else:
            return self.get_local_file_url(base_url, dataset_uid, obj_i, self._adata_path, self.local_file_uid)

    def get_ref_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._ref_url
        else:
            return self.get_local_file_url(base_url, dataset_uid, obj_i, self._ref_path, self.local_ref_uid)

    def make_file_def_creator(self, dataset_uid, obj_i):
        def get_anndata_zarr(base_url):
            options = {}
            options = gen_obs_locations_schema(self._spatial_centroid_obsm, options)
            options = gen_obs_segmentations_schema(self._spatial_polygon_obsm, options)
            options = gen_obs_spots_schema(self._spatial_spots_obsm, options)
            options = gen_obs_points_schema(self._spatial_points_obsm, options)
            options = gen_obs_embedding_schema(options, self._mappings_obsm, self._mappings_obsm_names, self._mappings_obsm_dims)
            options = gen_obs_sets_schema(options, self._obs_set_elems, self._obs_set_names,)
            options = gen_obs_feature_matrix_schema(options, self._expression_matrix, self._gene_var_filter, self._matrix_gene_var_filter)
            options = gen_feature_labels_schema(self._feature_labels, options)
            options = gen_obs_labels_schema(options, self._obs_labels_elems, self._obs_labels_names)

            if len(options.keys()) > 0:
                if self.is_h5ad:
                    options["refSpecUrl"] = self.get_ref_url(base_url, dataset_uid, obj_i)

                obj_file_def = {
                    "fileType": ft.ANNDATA_ZARR.value if not self.is_h5ad else ft.ANNDATA_H5AD.value,
                    "url": self.get_zarr_url(base_url, dataset_uid, obj_i) if not self.is_h5ad else self.get_h5ad_url(base_url, dataset_uid, obj_i),
                    "options": options
                }
                if self._request_init is not None:
                    obj_file_def['requestInit'] = self._request_init
                if self._coordination_values is not None:
                    obj_file_def['coordinationValues'] = self._coordination_values
                return obj_file_def
            return None
        return get_anndata_zarr

    def auto_view_config(self, vc):
        dataset = vc.add_dataset().add_object(self)
        mapping_name = self._mappings_obsm_names[0] if (
            self._mappings_obsm_names is not None) else self._mappings_obsm[0].split('/')[-1]
        scatterplot = vc.add_view(
            cm.SCATTERPLOT, dataset=dataset, mapping=mapping_name)
        cell_sets = vc.add_view(cm.OBS_SETS, dataset=dataset)
        genes = vc.add_view(cm.FEATURE_LIST, dataset=dataset)
        heatmap = vc.add_view(cm.HEATMAP, dataset=dataset)
        if self._spatial_polygon_obsm is not None or self._spatial_centroid_obsm is not None:
            spatial = vc.add_view(cm.SPATIAL, dataset=dataset)
            vc.layout((scatterplot | spatial)
                      / (heatmap | (cell_sets / genes)))
        else:
            vc.layout((scatterplot | (cell_sets / genes))
                      / heatmap)


SpatialDataWrapperType = TypeVar('SpatialDataWrapperType', bound='SpatialDataWrapper')


class SpatialDataWrapper(AnnDataWrapper):

    def __init__(self, sdata_path: Optional[str] = None, sdata_url: Optional[str] = None, sdata_store: Optional[Union[str, zarr.storage.StoreLike]] = None, sdata_artifact: Optional[ln.Artifact] = None, image_path: Optional[str] = None, region: Optional[str] = None, coordinate_system: Optional[str] = None, affine_transformation: Optional[np.ndarray] = None, obs_spots_path: Optional[str] = None, labels_path: Optional[str] = None, table_path: str = "tables/table", **kwargs):
        """
        Wrap a SpatialData object.

        :param sdata_path: SpatialData path, exclusive with other `{sdata,adata}_xxxx` arguments, by default None
        :type sdata_path: Optional[str]
        :param sdata_url: SpatialData url, exclusive with other `{sdata,adata}_xxxx` arguments, by default None
        :type sdata_url: Optional[str]
        :param sdata_store: SpatialData store, exclusive with other `{spatialdata,adata}_xxxx` arguments, by default None
        :type sdata_store: Optional[Union[str, zarr.storage.StoreLike]]
        :param sdata_artifact: Artifact that corresponds to a SpatialData object.
        :type sdata_artifact: Optional[ln.Artifact]
        :param image_path: Path to the image element of interest. By default, None.
        :type image_path: Optional[str]
        :param coordinate_system: Name of a target coordinate system.
        :type coordinate_system: Optional[str]
        :param affine_transformation: Transformation to be applied to the image. By default, None. Prefer coordinate_system.
        :type affine_transformation: Optional[np.ndarray]
        :param obs_spots_path: Location of shapes that should be interpreted as spot observations, by default None
        :type obs_spots_path: Optional[str]
        :param labels_path: Location of the labels (segmentation bitmask image), by default None
        :type labels_path: Optional[str]
        """
        raise_error_if_zero_or_more_than_one([
            sdata_path,
            sdata_url,
            sdata_store,
            sdata_artifact,
        ])
        raise_error_if_any([
            kwargs.get('adata_path', None),
            kwargs.get('adata_url', None),
            kwargs.get('adata_store', None),
            kwargs.get('adata_artifact', None)
        ])
        super().__init__(adata_path=sdata_path, adata_url=sdata_url, adata_store=sdata_store, adata_artifact=sdata_artifact, **kwargs)
        self.local_dir_uid = make_unique_filename(".sdata.zarr")
        self._image_path = image_path
        self._region = region
        self._coordinate_system = coordinate_system
        self._affine_transformation = affine_transformation
        self._kwargs = kwargs
        self._obs_spots_path = obs_spots_path
        self._labels_path = labels_path
        if self._adata_path is not None:
            self.zarr_folder = 'spatialdata.zarr'
        self.obs_type_label = None
        if self._coordination_values is not None and "obsType" in self._coordination_values:
            self.obs_type_label = self._coordination_values["obsType"]
        self._table_path = table_path

    @classmethod
    def from_object(cls: Type[SpatialDataWrapperType], sdata: SpatialData, table_keys_to_image_elems: dict[str, Union[str, None]] = defaultdict(type(None)), table_keys_to_regions: dict[str, Union[str, None]] = defaultdict(type(None)), obs_type_label: str = "spot") -> list[SpatialDataWrapperType]:
        """Instantiate a wrapper for SpatialData stores, one per table, directly from the SpatialData object.
        By default, we "show everything" that can reasonable be inferred given the information.  If you wish to have more control,
        consider instantiating the object directly.  This function will error if something cannot be inferred i.e., the user does not present
        regions explicitly but there is more than one for a given table.


        Parameters
        ----------
        cls : Type[SpatialDataWrapperType]
            _description_
        spatialdata : SpatialData
            _description_
        table_keys_to_image_elems : dict[str, str], optional
            which image paths to use for a given table for the visualization, by default None for each table key.
        table_keys_to_regions : dict[str, str], optional
            which regions to use for a given table for the visualization, by default None for each table key.

        Returns
        -------
        list[SpatialDataWrapperType]

        Raises
        ------
        ValueError
        """
        wrappers = []
        parent_table_key = "table" if (sdata.path / "table").exists() else "tables"
        for table_key, table in sdata.tables.items():
            spot_shapes_elem = None
            image_elem = table_keys_to_image_elems[table_key]
            labels_elem = None
            spatialdata_attr = table.uns['spatialdata_attrs']
            region = table_keys_to_regions[table_key]
            if region is not None:
                assert region in spatialdata_attr['region']
            else:
                region = spatialdata_attr['region']
            if isinstance(region, list):
                if len(region) > 1:
                    raise ValueError("Vitessce cannot subset AnnData objects on the fly.  Please provide an explicit region")
                region = region[0]
            if region in sdata.shapes:
                spot_shapes_elem = f"shapes/{region}"
                # Currently, only circle shapes are supported.
                # TODO: add if statement to check that this region contains spot shapes rather than other types of shapes
            if region in sdata.labels:
                labels_elem = f"labels/{region}"
            obs_feature_matrix_elem = f"{parent_table_key}/{table_key}/X"
            if 'highly_variable' in table.var:
                # TODO: fix first key needing to be "table" in vitessce-js
                initial_feature_filter_elem = 'highly_variable'
            else:
                initial_feature_filter_elem = None
            obs_set_elems = [f"{parent_table_key}/{table_key}/obs/{elem}" for elem in table.obs if table.obs[elem].dtype == 'category']
            wrappers += [
                cls(
                    sdata_path=str(sdata.path),
                    image_path=str(image_elem) if image_elem is not None else None,
                    labels_path=str(labels_elem) if labels_elem is not None else None,
                    obs_feature_matrix_path=str(obs_feature_matrix_elem),
                    obs_spots_path=str(spot_shapes_elem) if spot_shapes_elem is not None else None,
                    initial_feature_filter_path=initial_feature_filter_elem,
                    obs_set_paths=obs_set_elems,
                    coordination_values={"obsType": "spot"}  # TODO: should we remove?
                )
            ]
        return wrappers

    def make_file_def_creator(self, dataset_uid: str, obj_i: str) -> Optional[Callable]:
        def generator(base_url):
            options = {}
            options = gen_obs_labels_schema(options, self._obs_labels_elems, self._obs_labels_names)
            options = gen_sdata_obs_feature_matrix_schema(options, self._expression_matrix, self._gene_var_filter, self._matrix_gene_var_filter, self._region)
            options = gen_sdata_obs_sets_schema(options, self._obs_set_elems, self._obs_set_names, self._table_path, self._region)
            options = gen_sdata_obs_spots_schema(options, self._obs_spots_path, self._table_path, self._region, self._coordinate_system)
            options = gen_sdata_image_schema(options, self._image_path, self._coordinate_system, self._affine_transformation)
            options = gen_sdata_labels_schema(options, self._labels_path, self._table_path, self._coordinate_system, self._affine_transformation)
            options = gen_feature_labels_schema(self._feature_labels, options)
            if len(options.keys()) > 0:
                obj_file_def = {
                    "fileType": ft.SPATIALDATA_ZARR.value,
                    "url": self.get_zarr_url(base_url, dataset_uid, obj_i),
                    "options": options
                }
                if self._request_init is not None:
                    obj_file_def['requestInit'] = self._request_init
                if self._coordination_values is not None:
                    obj_file_def['coordinationValues'] = self._coordination_values
                return obj_file_def
            return None

        return generator


class MultivecZarrWrapper(AbstractWrapper):

    def __init__(self, zarr_path=None, zarr_url=None, **kwargs):
        super().__init__(**kwargs)
        self._repr = make_repr(locals())
        if zarr_url is not None and zarr_path is not None:
            raise ValueError(
                "Did not expect zarr_path to be provided with zarr_url")
        if zarr_url is None and zarr_path is None:
            raise ValueError(
                "Expected either zarr_url or zarr_path to be provided")
        self._zarr_path = zarr_path
        self._zarr_url = zarr_url
        if self._zarr_path is not None:
            self.is_remote = False
        else:
            self.is_remote = True
        self.local_dir_uid = make_unique_filename(".multivec.zarr")

    def convert_and_save(self, dataset_uid, obj_i, base_dir=None):
        # Only create out-directory if needed
        if not self.is_remote:
            super().convert_and_save(dataset_uid, obj_i, base_dir=base_dir)

        file_def_creator = self.make_genomic_profiles_file_def_creator(
            dataset_uid, obj_i)
        routes = self.make_genomic_profiles_routes(dataset_uid, obj_i)

        self.file_def_creators.append(file_def_creator)
        self.routes += routes

    def make_genomic_profiles_routes(self, dataset_uid, obj_i):
        if self.is_remote:
            return []
        else:
            return self.get_local_dir_route(dataset_uid, obj_i, self._zarr_path, self.local_dir_uid)

    def get_zarr_url(self, base_url="", dataset_uid="", obj_i=""):
        if self.is_remote:
            return self._zarr_url
        return self.get_local_dir_url(base_url, dataset_uid, obj_i, self._zarr_path, self.local_dir_uid)

    def make_genomic_profiles_file_def_creator(self, dataset_uid, obj_i):
        def genomic_profiles_file_def_creator(base_url):
            obj_file_def = {
                "fileType": "genomic-profiles.zarr",
                "url": self.get_zarr_url(base_url, dataset_uid, obj_i)
            }
            if self._request_init is not None:
                obj_file_def['requestInit'] = self._request_init
            return obj_file_def
        return genomic_profiles_file_def_creator



import pytest
from unittest.mock import patch, Mock
from copy import deepcopy

from vitessce import (
    CellBrowserToAnndataZarrConverter,
    convert_cell_browser_project_to_anndata,
)


valid_cellbrowser_config = {
    "fileVersions": {
        "inMatrix": {
            "fname": "/hive/data/inside/cells/datasets/adultPancreas/exprMatrix.csv.gz",
            "md5": "8da7a759a8",
            "size": 23363664,
            "mtime": "2018-10-16 23:29:40"
        },
        "outMatrix": {
            "fname": "/usr/local/apache/htdocs-cells/adultPancreas/exprMatrix.tsv.gz",
            "md5": "934bbdeacd",
            "size": 22710325,
            "mtime": "2022-05-18 22:34:06"
        },
        "inMeta": {
            "fname": "/hive/data/inside/cells/datasets/adultPancreas/meta.tsv",
            "md5": "7699cf188d",
            "size": 527639,
            "mtime": "2019-02-26 16:08:50"
        },
        "outMeta": {
            "fname": "/usr/local/apache/htdocs-cells/adultPancreas/meta.tsv",
            "md5": "cdfeda9e0a",
            "size": 522326,
            "mtime": "2022-05-24 18:01:35"
        },
    },
    "coords": [
        {
            "name": "coords_0",
            "shortLabel": "t-SNE",
            "md5": "3ff37334ef",
            "minX": 0,
            "maxX": 65535,
            "minY": 0,
            "maxY": 65535,
            "type": "Uint16",
            "textFname": "test.coords.tsv.gz",
            "labelMd5": "d41d8cd98f"
        }
    ],
    "topMarkers": {
        "acinar": [
            "A1CF",
        ],
        "alpha": [
            "A1BG-AS1",
        ],
        "beta": [
            "LEPR",
        ],
        "delta": [
            "SST",
            "RBP4",
        ],
        "ductal": [
            "ANXA4",
        ],
        "mesenchymal": [
            "SPARCL1",
        ],
        "nan": [
            "ERCC-00092",
        ],
        "unsure": [
            "G6PC2",
            "PCSK1",
        ]
    },
}

invalid_cellbrowser_config = deepcopy(valid_cellbrowser_config["fileVersions"])

coords_with_no_fname = {
    "coords": [
        {
            "name": "coords_0",
            "shortLabel": "t-SNE",
            "md5": "3ff37334ef",
            "minX": 0,
            "maxX": 65535,
            "minY": 0,
            "maxY": 65535,
            "type": "Uint16",
            "labelMd5": "d41d8cd98f"
        }
    ],
}

coords_with_no_fname_multi_word_shortlabel = {
    "coords": [
        {
            "name": "coords_0",
            "shortLabel": "Seurat umap",
            "md5": "3ff37334ef",
            "minX": 0,
            "maxX": 65535,
            "minY": 0,
            "maxY": 65535,
            "type": "Uint16",
            "labelMd5": "d41d8cd98f"
        }
    ],
}

cellbrowser_config_no_coords_filename = {
    **valid_cellbrowser_config,
    **coords_with_no_fname
}

cellbrowser_config_no_coords_filename_multi_word_shortlabel = {
    **valid_cellbrowser_config,
    **coords_with_no_fname_multi_word_shortlabel
}

project_name = "test-project"

"""
@pytest.fixture
def mock_makedirs():
    with patch('os.makedirs') as mock:
        yield mock


@pytest.fixture
def mock_write_zarr():
    with patch('anndata.AnnData.write_zarr') as mock:
        yield mock


@pytest.fixture
def mock_write_adata():
    with patch('anndata.AnnData.write') as mock:
        yield mock

@pytest.fixture
def mock_read_adata():
    with patch('anndata.read') as mock:
        # Create a dummy AnnData object
        dummy_data = np.random.rand(10, 10)
        mock_adata = anndata.AnnData(dummy_data)

        # Mock _read_adata to return the dummy AnnData object
        mock.return_value = mock_adata
        yield mock
"""


@pytest.fixture
def mock_filter_cells():
    with patch('scanpy.pp.filter_cells') as mock:
        yield mock


@pytest.fixture
def mock_end_to_end_tests(request):
    config = request.param
    mock_response_json = Mock()
    mock_response_json.json.return_value = config

    with open('tests/data/smaller_expr_matrix.tsv.gz', 'rb') as f:
        mock_response_expr_matrix = Mock()
        mock_response_expr_matrix.content = f.read()

    with open('tests/data/test_meta.tsv', 'rb') as f:
        mock_response_meta = Mock()
        mock_response_meta.content = f.read()

    with open('tests/data/test.coords.tsv.gz', 'rb') as f:
        mock_response_coords = Mock()
        mock_response_coords.content = f.read()

    with patch('requests.get') as mock_get:
        mock_get.side_effect = [mock_response_json, mock_response_expr_matrix, mock_response_meta, mock_response_coords]
        yield mock_get


def test_download_valid_config():

    with patch('requests.get') as mock_get:
        mock_get.return_value.json.return_value = valid_cellbrowser_config
        obj = CellBrowserToAnndataZarrConverter(project_name, keep_only_marker_genes=False)
        is_valid = obj.download_config()

        mock_get.assert_called_once_with('https://cells.ucsc.edu/test-project/dataset.json')
        assert is_valid
        assert obj.cellbrowser_config == valid_cellbrowser_config


@pytest.mark.parametrize('mock_end_to_end_tests', [valid_cellbrowser_config], indirect=True)
def test_filter_based_on_marker_genes(mock_end_to_end_tests, mock_filter_cells):

    inst = CellBrowserToAnndataZarrConverter(project_name, keep_only_marker_genes=True)
    config_is_valid = inst.download_config()

    assert config_is_valid
    inst.create_anndata_object()

    assert inst.adata.shape == (8, 1)

    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/dataset.json")
    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/exprMatrix.tsv.gz")
    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/meta.tsv")
    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/test.coords.tsv.gz")

    assert mock_end_to_end_tests.call_count == 4
    assert mock_filter_cells.call_count == 1


@pytest.mark.parametrize(
    'mock_end_to_end_tests, expected',
    [
        (valid_cellbrowser_config, "test.coords.tsv.gz"),
        (cellbrowser_config_no_coords_filename, "tMinusSNE.coords.tsv.gz"),
        (cellbrowser_config_no_coords_filename_multi_word_shortlabel, "Seurat_umap.coords.tsv.gz")
    ], indirect=["mock_end_to_end_tests"]
)
def test_end_to_end(mock_filter_cells, mock_end_to_end_tests, expected):
    convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/dataset.json")
    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/exprMatrix.tsv.gz")
    mock_end_to_end_tests.assert_any_call("https://cells.ucsc.edu/test-project/meta.tsv")
    mock_end_to_end_tests.assert_any_call(f"https://cells.ucsc.edu/test-project/{expected}")

    assert mock_end_to_end_tests.call_count == 4
    assert mock_filter_cells.call_count == 1


def test_end_to_end_invalid_config(mock_filter_cells):
    with patch('requests.get') as mock_get:
        mock_get.return_value.json.return_value = invalid_cellbrowser_config
        with pytest.raises(ValueError):
            convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

        mock_get.assert_called_once_with("https://cells.ucsc.edu/test-project/dataset.json")

    assert mock_get.call_count == 1
    assert mock_filter_cells.call_count == 0


def test_end_to_end_download_config_raises_exception(mock_filter_cells):
    mock_response = Mock()
    mock_response.raise_for_status.side_effect = Exception("Error downloading file")

    with patch('requests.get') as mock_get:
        mock_get.return_value = mock_response
        with pytest.raises(Exception):
            convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

        mock_get.assert_called_once_with("https://cells.ucsc.edu/test-project/dataset.json")

        assert mock_get.call_count == 1
    assert mock_filter_cells.call_count == 0


def test_end_to_end_load_expr_matrix_raises_exception(mock_filter_cells):
    mock_first_response = Mock()
    mock_first_response.json.return_value = valid_cellbrowser_config

    mock_second_response = Mock()
    mock_second_response.raise_for_status.side_effect = Exception("Error downloading file")

    with patch('requests.get') as mock_get:
        mock_get.side_effect = [mock_first_response, mock_second_response]
        with pytest.raises(Exception):
            convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/dataset.json")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/exprMatrix.tsv.gz")
        assert mock_get.call_count == 2
    assert mock_filter_cells.call_count == 0


def test_end_to_end_load_cell_metadata_raises_exception(mock_filter_cells):
    mock_get_config = Mock()
    mock_get_config.json.return_value = valid_cellbrowser_config

    with open('tests/data/smaller_expr_matrix.tsv.gz', 'rb') as f:
        mock_response_expr_matrix = Mock()
        mock_response_expr_matrix.content = f.read()
        mock_response_expr_matrix.raise_for_status.return_value = None

    mock_response_meta = Mock()
    mock_response_meta.raise_for_status.side_effect = Exception("Error downloading file")
    assert mock_filter_cells.call_count == 0

    with patch('requests.get') as mock_get:
        mock_get.side_effect = [mock_get_config, mock_response_expr_matrix, mock_response_meta]
        with pytest.raises(Exception):
            convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/dataset.json")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/exprMatrix.tsv.gz")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/meta.tsv")
        assert mock_get.call_count == 3
    assert mock_filter_cells.call_count == 0


def test_end_to_end_add_coords_raises_exception(mock_filter_cells):
    mock_get_config = Mock()
    mock_get_config.json.return_value = valid_cellbrowser_config

    with open('tests/data/smaller_expr_matrix.tsv.gz', 'rb') as f:
        mock_response_expr_matrix = Mock()
        mock_response_expr_matrix.content = f.read()
        mock_response_expr_matrix.raise_for_status.return_value = None

    with open('tests/data/test_meta.tsv', 'rb') as f:
        mock_response_meta = Mock()
        mock_response_meta.content = f.read()
        mock_response_meta.raise_for_status.return_value = None

    mock_coords = Mock()
    mock_coords.raise_for_status.side_effect = Exception("Error downloading file")

    with patch('requests.get') as mock_get:
        mock_get.side_effect = [mock_get_config, mock_response_expr_matrix, mock_response_meta, mock_coords]
        with pytest.raises(Exception):
            convert_cell_browser_project_to_anndata(project_name, keep_only_marker_genes=False)

        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/dataset.json")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/exprMatrix.tsv.gz")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/meta.tsv")
        mock_get.assert_any_call("https://cells.ucsc.edu/test-project/test.coords.tsv.gz")
        assert mock_get.call_count == 4
    assert mock_filter_cells.call_count == 0



from esbuild_py import transform
from ..widget import VitesscePlugin


PLUGIN_ESM = transform("""
function createPlugins(utilsForPlugins) {
    const {
        React,
        PluginFileType,
        PluginViewType,
        PluginCoordinationType,
        PluginJointFileType,
        z,
        useCoordination,
    } = utilsForPlugins;
    function SpatialQueryView(props) {
        const { coordinationScopes } = props;
        const [{
            queryParams,
            obsSetSelection,
        }, {
            setQueryParams,
        }] = useCoordination(['queryParams', 'obsSetSelection', 'obsType'], coordinationScopes);

        const [uuid, setUuid] = React.useState(1);
        const [queryType, setQueryType] = React.useState('grid');
        const [maxDist, setMaxDist] = React.useState(100);
        const [minSize, setMinSize] = React.useState(4);
        const [minCount, setMinCount] = React.useState(10);
        const [minSupport, setMinSupport] = React.useState(0.5);

        const cellTypeOfInterest = obsSetSelection?.length === 1 && obsSetSelection[0][0] === "Cell Type"
            ? obsSetSelection[0][1]
            : null;

        const onQueryTypeChange = React.useCallback((e) => {
            setQueryType(e.target.value);
        }, []);

        return (
        <div className="spatial-query">
            <p>Spatial Query Manager</p>
            <label>
                Query type&nbsp;
                <select onChange={onQueryTypeChange}>
                    <option value="grid">Grid-based</option>
                    <option value="rand">Random-based</option>
                    <option value="ct-center" disabled={cellTypeOfInterest === null}>Cell type of interest</option>
                </select>
            </label>
            <br/>
            <label>
                {/* Maximum distance to consider a cell as a neighbor. */}
                Max. Dist.
                <input type="range" value={maxDist} onChange={e => setMaxDist(parseFloat(e.target.value))} min={50} max={250} step={1} />
                {maxDist}
            </label>
            <br/>
            <label>
                {/* Minimum neighborhood size for each point to consider. */}
                Min. Size
                <input type="range" value={minSize} onChange={e => setMinSize(parseFloat(e.target.value))} min={0} max={20} step={1} />
                {minSize}
            </label>
            <br/>
            <label>
                {/* Minimum number of cell type to consider. */}
                Min. Count
                <input type="range" value={minCount} onChange={e => setMinCount(parseFloat(e.target.value))} min={0} max={30} step={1} />
                {minCount}
            </label>
            <br/>
            <label>
                {/* Threshold of frequency to consider a pattern as a frequent pattern. */}
                Min. Support
                <input type="range" value={minSupport} onChange={e => setMinSupport(parseFloat(e.target.value))} min={0} max={1} step={0.01} />
                {minSupport}
            </label>
            <br/>
            {/* TODO: disDuplicates: Distinguish duplicates in patterns. */}
            <button onClick={(e) => {
                setQueryParams({
                    cellTypeOfInterest,
                    queryType,
                    maxDist,
                    minSize,
                    minCount,
                    minSupport,
                    uuid,
                });
                setUuid(uuid+1);
            }}>Find patterns</button>
        </div>
        );
    }

    const pluginCoordinationTypes = [
        new PluginCoordinationType('queryParams', null, z.object({
            cellTypeOfInterest: z.string().nullable(),
            queryType: z.enum(['grid', 'rand', 'ct-center']),
            maxDist: z.number(),
            minSize: z.number(),
            minCount: z.number(),
            minSupport: z.number(),
            disDuplicates: z.boolean(),
            uuid: z.number(),
        }).partial().nullable()),
    ];

    const pluginViewTypes = [
        new PluginViewType('spatialQuery', SpatialQueryView, ['queryParams', 'obsSetSelection', 'obsType']),
    ];
    return { pluginViewTypes, pluginCoordinationTypes };
}
export default { createPlugins };
""")


class SpatialQueryPlugin(VitesscePlugin):
    """
    Spatial-Query plugin view renders controls to change parameters passed to the Spatial-Query methods.
    """
    plugin_esm = PLUGIN_ESM
    commands = {}

    def __init__(self, adata, spatial_key="X_spatial", label_key="cell_type"):
        """
        Construct a new Vitessce widget.

        :param adata: AnnData.
        :type adata: anndata.AnnData
        :param str spatial_key: The key in adata.obsm that contains the (x, y) coordinates of each cell. By default, "X_spatial".
        :param str label_key: The column in adata.obs that contains the cell type labels. By default, "cell_type".

        .. code-block:: python

            from vitessce.widget_plugins import SpatialQueryPlugin

            plugin = SpatialQueryPlugin(adata, spatial_key="X_spatial", label_key="cell_type")
            # ...
            vc.widget(plugins=[plugin], remount_on_uid_change=False)
        """
        from SpatialQuery.spatial_query import spatial_query
        import matplotlib.pyplot as plt  # Add as dependency / optional dependency?

        self.adata = adata
        self.spatial_key = spatial_key
        self.label_key = label_key

        self.tt = spatial_query(adata=adata, dataset='test', spatial_key=spatial_key, label_key=label_key, leaf_size=10)

        self.tab20_rgb = [[int(r * 255), int(g * 255), int(b * 255)] for (r, g, b, a) in [plt.cm.tab20(i) for i in range(20)]]

        self.additional_obs_sets = {
            "version": "0.1.3",
            "tree": [
                {
                    "name": "Spatial-Query Results",
                    "children": [

                    ]
                }
            ]
        }

        self.obs_set_color = [
            {
                "color": [255, 255, 255],
                "path": ["Cell Type"],
            },
            {
                "color": [255, 255, 255],
                "path": ["Spatial-Query Results"],
            }
        ]

        self.ct_to_color = dict()

        for ct_i, cell_type in enumerate(adata.obs[label_key].unique().tolist()):
            color = self.tab20_rgb[ct_i % 20]
            self.ct_to_color[cell_type] = color
            path = ["Cell Type", cell_type]
            self.obs_set_color.append({
                "color": color,
                "path": path
            })

        self.cell_i_to_cell_id = dict(zip(range(adata.obs.shape[0]), adata.obs.index.tolist()))
        self.cell_id_to_cell_type = dict(zip(adata.obs.index.tolist(), adata.obs[label_key].tolist()))

    def get_matching_cell_ids(self, cell_type, cell_i):
        cell_ids = [self.cell_i_to_cell_id[i] for i in cell_i]
        matches = []
        for cell_id in cell_ids:
            cell_ct = self.cell_id_to_cell_type[cell_id]
            if cell_ct == cell_type:
                matches.append([cell_id, None])
        return matches

    def fp_tree_to_obs_sets_tree(self, fp_tree, sq_id):
        additional_obs_sets = {
            "version": "0.1.3",
            "tree": [
                {
                    "name": f"Spatial-Query Results {sq_id}",
                    "children": [

                    ]
                }
            ]
        }

        obs_set_color = []

        for row_i, row in fp_tree.iterrows():
            try:
                motif = row["itemsets"]
            except KeyError:
                motif = row["motifs"]
            cell_i = row["cell_id"]

            motif_name = str(list(motif))

            additional_obs_sets["tree"][0]["children"].append({
                "name": motif_name,
                "children": [
                    {
                        "name": cell_type,
                        "set": self.get_matching_cell_ids(cell_type, cell_i)
                    }
                    for cell_type in motif
                ]
            })

            obs_set_color.append({
                "color": [255, 255, 255],
                "path": [additional_obs_sets["tree"][0]["name"], motif_name]
            })

            for cell_type in motif:
                color = self.ct_to_color[cell_type]
                path = [additional_obs_sets["tree"][0]["name"], motif_name, cell_type]
                obs_set_color.append({
                    "color": color,
                    "path": path
                })
        return (additional_obs_sets, obs_set_color)

    def run_sq(self, prev_config):
        query_params = prev_config["coordinationSpace"]["queryParams"]["A"]

        max_dist = query_params.get("maxDist", 150)
        min_size = query_params.get("minSize", 4)
        # min_count = query_params.get("minCount", 10)
        min_support = query_params.get("minSupport", 0.5)
        # dis_duplicates = query_params.get("disDuplicates", False)  # if distinguish duplicates of cell types in neighborhood
        query_type = query_params.get("queryType", "grid")
        cell_type_of_interest = query_params.get("cellTypeOfInterest", None)

        query_uuid = query_params["uuid"]

        params_dict = dict(
            max_dist=max_dist,
            min_size=min_size,
            # min_count=min_count,
            min_support=min_support,
            # dis_duplicates=dis_duplicates,
            if_display=True,
            fig_size=(9, 6),
            return_cellID=True,
        )
        print(params_dict)

        # TODO: add unit tests for this functionality

        if query_type == "rand":
            # TODO: implement param similar to return_grid for find_patterns_rand (to return the random points used)
            fp_tree = self.tt.find_patterns_rand(**params_dict)
        elif query_type == "grid":
            params_dict["return_grid"] = True
            fp_tree, grid_pos = self.tt.find_patterns_grid(**params_dict)
        elif query_type == "ct-center":
            fp_tree = self.tt.motif_enrichment_knn(
                ct=cell_type_of_interest,
                k=20,  # TODO: make this a parameter in the UI.
                min_support=min_support,
                # dis_duplicates=dis_duplicates,
                return_cellID=True,
            )
            print(fp_tree)

        # TODO: implement query types that are dependent on motif selection.

        # Previous values
        additional_obs_sets = prev_config["coordinationSpace"]["additionalObsSets"]["A"]
        obs_set_color = prev_config["coordinationSpace"]["obsSetColor"]["A"]

        # Perform query
        (new_additional_obs_sets, new_obs_set_color) = self.fp_tree_to_obs_sets_tree(fp_tree, query_uuid)

        additional_obs_sets["tree"][0] = new_additional_obs_sets["tree"][0]
        prev_config["coordinationSpace"]["additionalObsSets"]["A"] = additional_obs_sets

        obs_set_color += new_obs_set_color
        prev_config["coordinationSpace"]["obsSetColor"]["A"] = obs_set_color

        motif_to_select = new_additional_obs_sets["tree"][0]["children"][0]["name"]
        new_obs_set_selection = [[new_additional_obs_sets["tree"][0]["name"], motif_to_select, node["name"]] for node in new_additional_obs_sets["tree"][0]["children"][0]["children"]]
        prev_config["coordinationSpace"]["obsSetSelection"]["A"] = new_obs_set_selection

        # TODO: need to fix bug that prevents this from working
        # Reference: https://github.com/vitessce/vitessce/blob/774328ab5c4436576dd2e8e4fff0758d6c6cce89/packages/view-types/obs-sets-manager/src/ObsSetsManagerSubscriber.js#L104
        prev_config["coordinationSpace"]["obsSetExpansion"]["A"] = [path[:-1] for path in new_obs_set_selection]

        return {**prev_config, "uid": f"with_query_{query_uuid}"}

    def on_config_change(self, new_config):
        query_params = new_config["coordinationSpace"]["queryParams"]["A"]
        if query_params and "uuid" in query_params:
            print(query_params)
            query_uuid = query_params.get("uuid", None)
            if new_config["uid"] != f"with_query_{query_uuid}":
                return self.run_sq(new_config)
        return None
